"""NumPyro-specific conversion code."""

import warnings
from collections import defaultdict

import numpy as np
from xarray import DataTree

from arviz_base.base import dict_to_dataset, requires
from arviz_base.rcparams import rc_context, rcParams
from arviz_base.utils import expand_dims


class SVIWrapper:
    """A helper class for SVI to mimic numpyro.infer.MCMC methods."""

    def __init__(
        self,
        svi,
        *,
        svi_result,
        model_args=None,
        model_kwargs=None,
        num_samples: int = 1000,
    ):
        import jax
        import numpyro

        self.svi = svi
        self.svi_result = svi_result
        self._args = model_args or tuple()
        self._kwargs = model_kwargs or dict()
        self.num_samples = num_samples
        self.thinning = 1
        self.num_chains = 0
        self.sample_dims = ["sample"]
        self.kind = "svi"

        self.numpyro = numpyro
        self.prng_key_func = jax.random.PRNGKey

    def get_samples(self, seed=None, **kwargs):
        """Mimics mcmc.get_samples()."""
        key = self.prng_key_func(seed or 0)
        if isinstance(self.svi.guide, self.numpyro.infer.autoguide.AutoGuide):
            return self.svi.guide.sample_posterior(
                key,
                self.svi_result.params,
                *self._args,
                sample_shape=(self.num_samples,),
                **self._kwargs,
            )
        # if a custom guide is provided, sample by hand
        predictive = self.numpyro.infer.Predictive(
            self.svi.guide, params=self.svi_result.params, num_samples=self.num_samples
        )
        samples = predictive(key, *self._args, **self._kwargs)
        return samples

    @property
    def sampler(self):
        """Mimics mcmc.sampler.model."""

        class Sampler:
            def __init__(self, model):
                self._model = model

            @property
            def model(self):
                return self._model

        return Sampler(getattr(self.svi.guide, "model", self.svi.model))

    def get_extra_fields(self, **kwargs):
        """Mimics mcmc.get_extra_fields()."""
        return dict()


def _add_dims(dims_a, dims_b):
    """Merge two dimension mappings by concatenating dimension labels.

    Used to combine batch dims with event dims by appending the dims of dims_b to dims_a.

    Parameters
    ----------
    dims_a : dict of {str: list of str(s)}
        Mapping from site name to a list of dimension labels, typically
        representing batch dimensions.
    dims_b : dict of {str: list of str(s)}
        Mapping from site name to a list of dimension labels, typically
        representing event dimensions.

    Returns
    -------
    dict of {str: list of str(s)}
        Combined mapping where each site name is associated with the
        concatenated dimension labels from both inputs.
    """
    merged = defaultdict(list, dims_a)
    for k, v in dims_b.items():
        merged[k].extend(v)

    # Convert back to a regular dict
    return dict(merged)


def infer_dims(
    model,
    model_args=None,
    model_kwargs=None,
):
    """Infers batch dim names from numpyro model plates.

    Parameters
    ----------
    model : callable
        A numpyro model function.
    model_args : tuple of (Any, ...), optional
        Input args for the numpyro model.
    model_kwargs : dict of {str: Any}, optional
        Input kwargs for the numpyro model.

    Returns
    -------
    dict of {str: list of str(s)}
        Mapping from model site name to list of dimension labels.
    """
    import jax
    from numpyro import distributions as dist
    from numpyro import handlers
    from numpyro.infer.initialization import init_to_sample
    from numpyro.ops.pytree import PytreeTrace

    model_args = tuple() if model_args is None else model_args
    model_kwargs = dict() if model_kwargs is None else model_kwargs

    def _get_dist_name(fn):
        if isinstance(fn, dist.Independent | dist.ExpandedDistribution | dist.MaskedDistribution):
            return _get_dist_name(fn.base_dist)
        return type(fn).__name__

    def get_trace():
        # We use `init_to_sample` to get around ImproperUniform distribution,
        # which does not have `sample` method.
        subs_model = handlers.substitute(
            handlers.seed(model, 0),
            substitute_fn=init_to_sample,
        )
        trace = handlers.trace(subs_model).get_trace(*model_args, **model_kwargs)
        # Work around an issue where jax.eval_shape does not work
        # for distribution output (e.g. the function `lambda: dist.Normal(0, 1)`)
        # Here we will remove `fn` and store its name in the trace.
        for _, site in trace.items():
            if site["type"] == "sample":
                site["fn_name"] = _get_dist_name(site.pop("fn"))
            elif site["type"] == "deterministic":
                site["fn_name"] = "Deterministic"
        return PytreeTrace(trace)

    # We use eval_shape to avoid any array computation.
    trace = jax.eval_shape(get_trace).trace

    named_dims = {}

    # loop through the trace and pull the batch dim and event dim names
    for name, site in trace.items():
        batch_dims = [frame.name for frame in sorted(site["cond_indep_stack"], key=lambda x: x.dim)]
        event_dims = list(site.get("infer", {}).get("event_dims", []))

        # save the dim names leading with batch dims
        if site["type"] in ["sample", "deterministic"] and (batch_dims or event_dims):
            named_dims[name] = batch_dims + event_dims

    return named_dims


class NumPyroConverter:
    """Encapsulate NumPyro specific logic."""

    # pylint: disable=too-many-instance-attributes

    model = None
    nchains = None
    ndraws = None

    def __init__(
        self,
        *,
        posterior=None,
        prior=None,
        posterior_predictive=None,
        predictions=None,
        constant_data=None,
        predictions_constant_data=None,
        log_likelihood=False,
        index_origin=None,
        coords=None,
        dims=None,
        pred_dims=None,
        extra_event_dims=None,
        num_chains=1,
    ):
        """Convert NumPyro data into an InferenceData object.

        Parameters
        ----------
        posterior : numpyro.mcmc.MCMC
            Fitted MCMC object from NumPyro
        prior : dict, optional
            Prior samples from a NumPyro model
        posterior_predictive : dict, optional
            Posterior predictive samples for the posterior
        predictions : dict, optional
            Out of sample predictions
        constant_data : dict, optional
            Dictionary containing constant data variables mapped to their values.
        predictions_constant_data : dict, optional
            Constant data used for out-of-sample predictions.
        index_origin : int, optional
        coords : dict, optional
            Map of dimensions to coordinates
        dims : dict of {str : list of str}, optional
            Map variable names to their coordinates. Will be inferred if they are not provided.
        pred_dims : dict, optional
            Dims for predictions data. Map variable names to their coordinates.
        num_chains : int, optional
            Number of chains used for sampling. Ignored if posterior is present.
        extra_event_dims : dict, optional
            Maps event dims that couldnt be inferred (ie deterministic sites) to their coordinates.
        """
        import jax
        import numpyro

        self.posterior = posterior
        self.prior = jax.device_get(prior)
        self.posterior_predictive = jax.device_get(posterior_predictive)
        self.predictions = predictions
        self.constant_data = constant_data
        self.predictions_constant_data = predictions_constant_data
        self.log_likelihood = log_likelihood
        self.index_origin = rcParams["data.index_origin"] if index_origin is None else index_origin
        self.coords = coords
        self.dims = dims
        self.pred_dims = pred_dims
        self.extra_event_dims = extra_event_dims
        self.numpyro = numpyro

        def arbitrary_element(dct):
            return next(iter(dct.values()))

        if posterior is not None:
            samples = jax.device_get(self.posterior.get_samples(group_by_chain=True))
            if hasattr(samples, "_asdict"):
                # In case it is easy to convert to a dictionary, as in the case of namedtuples
                samples = {k: expand_dims(v) for k, v in samples._asdict().items()}
            if not isinstance(samples, dict):
                # handle the case we run MCMC with a general potential_fn
                # (instead of a NumPyro model) whose args is not a dictionary
                # (e.g. f(x) = x ** 2)
                tree_flatten_samples = jax.tree_util.tree_flatten(samples)[0]
                samples = {
                    f"Param:{i}": jax.device_get(v) for i, v in enumerate(tree_flatten_samples)
                }
            self._samples = samples
            self.nchains, self.ndraws = (
                posterior.num_chains,
                posterior.num_samples // posterior.thinning,
            )
            self.model = self.posterior.sampler.model
            # model arguments and keyword arguments
            self._args = self.posterior._args  # pylint: disable=protected-access
            self._kwargs = self.posterior._kwargs  # pylint: disable=protected-access
            self.dims = self.dims if self.dims is not None else self.infer_dims()
            self.pred_dims = (
                self.pred_dims if self.pred_dims is not None else self.infer_pred_dims()
            )
        else:
            self.nchains = num_chains
            get_from = None
            if predictions is not None:
                get_from = predictions
            elif posterior_predictive is not None:
                get_from = posterior_predictive
            elif prior is not None:
                get_from = prior
            if get_from is None and constant_data is None and predictions_constant_data is None:
                raise ValueError(
                    "When constructing InferenceData must have at least"
                    " one of posterior, prior, posterior_predictive or predictions."
                )
            if get_from is not None:
                aelem = arbitrary_element(get_from)
                self.ndraws = aelem.shape[0] // self.nchains

        observations = {}
        if self.model is not None:
            trace = self._get_model_trace(
                self.model,
                model_args=self._args,
                model_kwargs=self._kwargs,
                key=jax.random.PRNGKey(0),
            )
            observations = {
                name: site["value"]
                for name, site in trace.items()
                if site["type"] == "sample" and site["is_observed"]
            }
        self.observations = observations if observations else None

    def _get_model_trace(self, model, model_args, model_kwargs, key):
        """Extract the numpyro model trace."""
        model_args = model_args or tuple()
        model_kwargs = model_kwargs or dict()

        # we need to use an init strategy to generate random samples for ImproperUniform sites
        seeded_model = self.numpyro.handlers.substitute(
            self.numpyro.handlers.seed(model, key),
            substitute_fn=self.numpyro.infer.init_to_sample,
        )
        trace = self.numpyro.handlers.trace(seeded_model).get_trace(*model_args, **model_kwargs)
        return trace

    @requires("posterior")
    def posterior_to_xarray(self):
        """Convert the posterior to an xarray dataset."""
        data = self._samples
        return dict_to_dataset(
            data,
            inference_library=self.numpyro,
            coords=self.coords,
            dims=self.dims,
            index_origin=self.index_origin,
        )

    @requires("posterior")
    def sample_stats_to_xarray(self):
        """Extract sample_stats from NumPyro posterior."""
        rename_key = {
            "potential_energy": "lp",
            "adapt_state.step_size": "step_size",
            "num_steps": "n_steps",
            "accept_prob": "acceptance_rate",
        }
        data = {}
        for stat, value in self.posterior.get_extra_fields(group_by_chain=True).items():
            if isinstance(value, dict | tuple):
                continue
            name = rename_key.get(stat, stat)
            value_cp = value.copy()
            data[name] = value_cp
            if stat == "num_steps":
                data["tree_depth"] = np.log2(value_cp).astype(int) + 1

        return dict_to_dataset(
            data,
            inference_library=self.numpyro,
            dims=None,
            coords=self.coords,
            index_origin=self.index_origin,
        )

    @requires("posterior")
    @requires("model")
    def log_likelihood_to_xarray(self):
        """Extract log likelihood from NumPyro posterior."""
        if not self.log_likelihood:
            return None
        data = {}
        if self.observations is not None:
            samples = self.posterior.get_samples(group_by_chain=False)
            if hasattr(samples, "_asdict"):
                samples = samples._asdict()
            log_likelihood_dict = self.numpyro.infer.log_likelihood(
                self.model, samples, *self._args, **self._kwargs
            )
            for obs_name, log_like in log_likelihood_dict.items():
                shape = (self.nchains, self.ndraws) + log_like.shape[1:]
                data[obs_name] = np.reshape(np.asarray(log_like), shape)
        return dict_to_dataset(
            data,
            inference_library=self.numpyro,
            dims=self.dims,
            coords=self.coords,
            index_origin=self.index_origin,
            skip_event_dims=True,
        )

    def translate_posterior_predictive_dict_to_xarray(self, dct, dims):
        """Convert posterior_predictive or prediction samples to xarray."""
        data = {}
        for k, ary in dct.items():
            shape = ary.shape
            if shape[0] == self.nchains and shape[1] == self.ndraws:
                data[k] = ary
            elif (
                shape[0] == self.nchains * self.ndraws
                and getattr(self.posterior, "kind", "") != "svi"
            ):
                data[k] = ary.reshape((self.nchains, self.ndraws, *shape[1:]))
            elif getattr(self.posterior, "kind", "") == "svi":
                data[k] = ary
            else:
                data[k] = expand_dims(ary)
                warnings.warn(
                    "posterior predictive shape not compatible with number of chains and draws. "
                    "This can mean that some draws or even whole chains are not represented."
                )
        return dict_to_dataset(
            data,
            inference_library=self.numpyro,
            coords=self.coords,
            dims=dims,
            index_origin=self.index_origin,
        )

    @requires("posterior_predictive")
    def posterior_predictive_to_xarray(self):
        """Convert posterior_predictive samples to xarray."""
        return self.translate_posterior_predictive_dict_to_xarray(
            self.posterior_predictive, self.dims
        )

    @requires("predictions")
    def predictions_to_xarray(self):
        """Convert predictions to xarray."""
        return self.translate_posterior_predictive_dict_to_xarray(self.predictions, self.pred_dims)

    def priors_to_xarray(self):
        """Convert prior samples (and if possible prior predictive too) to xarray."""
        if self.prior is None:
            return {"prior": None, "prior_predictive": None}
        if self.posterior is not None:
            prior_vars = list(self._samples.keys())
            prior_predictive_vars = [key for key in self.prior.keys() if key not in prior_vars]
        else:
            prior_vars = self.prior.keys()
            prior_predictive_vars = None

        # dont expand dims for SVI
        expand_dims_func = (
            expand_dims if getattr(self.posterior, "kind", "") != "svi" else lambda x: x
        )
        priors_dict = {
            group: (
                None
                if var_names is None
                else dict_to_dataset(
                    {k: expand_dims_func(self.prior[k]) for k in var_names},
                    inference_library=self.numpyro,
                    coords=self.coords,
                    dims=self.dims,
                    index_origin=self.index_origin,
                )
            )
            for group, var_names in zip(
                ("prior", "prior_predictive"), (prior_vars, prior_predictive_vars)
            )
        }
        return priors_dict

    @requires("observations")
    @requires("model")
    def observed_data_to_xarray(self):
        """Convert observed data to xarray."""
        return dict_to_dataset(
            self.observations,
            inference_library=self.numpyro,
            dims=self.dims,
            coords=self.coords,
            sample_dims=[],
            index_origin=self.index_origin,
        )

    @requires("constant_data")
    def constant_data_to_xarray(self):
        """Convert constant_data to xarray."""
        return dict_to_dataset(
            self.constant_data,
            inference_library=self.numpyro,
            dims=self.dims,
            coords=self.coords,
            sample_dims=[],
            index_origin=self.index_origin,
        )

    @requires("predictions_constant_data")
    def predictions_constant_data_to_xarray(self):
        """Convert predictions_constant_data to xarray."""
        return dict_to_dataset(
            self.predictions_constant_data,
            inference_library=self.numpyro,
            dims=self.pred_dims,
            coords=self.coords,
            sample_dims=[],
            index_origin=self.index_origin,
        )

    def to_datatree(self):
        """Convert all available data to an InferenceData object.

        Note that if groups can not be created (i.e., there is no `trace`, so
        the `posterior` and `sample_stats` can not be extracted), then the InferenceData
        will not have those groups.
        """
        dicto = {
            "posterior": self.posterior_to_xarray(),
            "sample_stats": self.sample_stats_to_xarray(),
            "log_likelihood": self.log_likelihood_to_xarray(),
            "posterior_predictive": self.posterior_predictive_to_xarray(),
            "predictions": self.predictions_to_xarray(),
            **self.priors_to_xarray(),
            "observed_data": self.observed_data_to_xarray(),
            "constant_data": self.constant_data_to_xarray(),
            "predictions_constant_data": self.predictions_constant_data_to_xarray(),
        }

        return DataTree.from_dict({group: ds for group, ds in dicto.items() if ds is not None})

    @requires("posterior")
    @requires("model")
    def infer_dims(self) -> dict[str, list[str]]:
        """Infers dims for input data."""
        dims = infer_dims(self.model, self._args, self._kwargs)
        if self.extra_event_dims:
            dims = _add_dims(dims, self.extra_event_dims)
        return dims

    @requires("posterior")
    @requires("model")
    @requires("predictions")
    def infer_pred_dims(self) -> dict[str, list[str]]:
        """Infers dims for predictions data."""
        dims = infer_dims(self.model, self._args, self._kwargs)
        if self.extra_event_dims:
            dims = _add_dims(dims, self.extra_event_dims)
        return dims


def from_numpyro(
    posterior=None,
    *,
    prior=None,
    posterior_predictive=None,
    predictions=None,
    constant_data=None,
    predictions_constant_data=None,
    log_likelihood=None,
    index_origin=None,
    coords=None,
    dims=None,
    pred_dims=None,
    extra_event_dims=None,
    num_chains=1,
):
    """Convert NumPyro data into a DataTree object.

    For a usage example read :ref:`numpyro_conversion`

    If no dims are provided, this will infer batch dim names from NumPyro model plates.
    For event dim names, such as with the ZeroSumNormal, `infer={"event_dims":dim_names}`
    can be provided in numpyro.sample, i.e.::

        # equivalent to dims entry, {"gamma": ["groups"]}
        gamma = numpyro.sample(
            "gamma",
            dist.ZeroSumNormal(1, event_shape=(n_groups,)),
            infer={"event_dims":["groups"]}
        )

    There is also an additional `extra_event_dims` input to cover any edge cases, for instance
    deterministic sites with event dims (which dont have an `infer` argument to provide metadata).

    Parameters
    ----------
    posterior : numpyro.mcmc.MCMC
        Fitted MCMC object from NumPyro
    prior : dict, optional
        Prior samples from a NumPyro model
    posterior_predictive : dict, optional
        Posterior predictive samples for the posterior
    predictions : dict, optional
        Out of sample predictions
    constant_data : dict, optional
        Dictionary containing constant data variables mapped to their values.
    predictions_constant_data : dict, optional
        Constant data used for out-of-sample predictions.
    index_origin : int, optional
    coords : dict, optional
        Map of dimensions to coordinates
    dims : dict of {str : list of str}, optional
        Map variable names to their coordinates. Will be inferred if they are not provided.
    pred_dims : dict, optional
        Dims for predictions data. Map variable names to their coordinates. Default behavior is to
        infer dims if this is not provided
    extra_event_dims : dict, optional
        Extra event dims for deterministic sites. Maps event dims that couldnt be inferred to
        their coordinates.
    num_chains : int, default 1
        Number of chains used for sampling. Ignored if posterior is present.

    Returns
    -------
    DataTree
    """
    with rc_context(rc={"data.sample_dims": ["chain", "draw"]}):
        return NumPyroConverter(
            posterior=posterior,
            prior=prior,
            posterior_predictive=posterior_predictive,
            predictions=predictions,
            constant_data=constant_data,
            predictions_constant_data=predictions_constant_data,
            log_likelihood=log_likelihood,
            index_origin=index_origin,
            coords=coords,
            dims=dims,
            pred_dims=pred_dims,
            extra_event_dims=extra_event_dims,
            num_chains=num_chains,
        ).to_datatree()


def from_numpyro_svi(
    svi,
    *,
    svi_result,
    model_args=None,
    model_kwargs=None,
    prior=None,
    posterior_predictive=None,
    predictions=None,
    constant_data=None,
    predictions_constant_data=None,
    log_likelihood=None,
    index_origin=None,
    coords=None,
    dims=None,
    pred_dims=None,
    extra_event_dims=None,
    model=None,
    num_samples: int = 1000,
):
    """Convert NumPyro SVI results into a DataTree object.

    For a usage example read :ref:`numpyro_conversion`

    If no dims are provided, this will infer batch dim names from NumPyro model plates.
    For event dim names, such as with the ZeroSumNormal, `infer={"event_dims":dim_names}`
    can be provided in numpyro.sample, i.e.::

        # equivalent to dims entry, {"gamma": ["groups"]}
        gamma = numpyro.sample(
            "gamma",
            dist.ZeroSumNormal(1, event_shape=(n_groups,)),
            infer={"event_dims":["groups"]}
        )

    There is also an additional `extra_event_dims` input to cover any edge cases, for instance
    deterministic sites with event dims (which dont have an `infer` argument to provide metadata).

    Parameters
    ----------
    svi : numpyro.infer.svi.SVI
        Numpyro SVI instance used for fitting the model.
    svi_result : numpyro.infer.svi.SVIRunResult
        SVI results from a fitted model.
    model_args : tuple, optional
        Model arguments, should match those used for fitting the model.
    model_kwargs : dict, optional
        Model keyword arguments, should match those used for fitting the model.
    prior : dict, optional
        Prior samples from a NumPyro model
    posterior_predictive : dict, optional
        Posterior predictive samples for the posterior
    predictions : dict, optional
        Out of sample predictions
    constant_data : dict, optional
        Dictionary containing constant data variables mapped to their values.
    predictions_constant_data : dict, optional
        Constant data used for out-of-sample predictions.
    index_origin : int, optional
    coords : dict, optional
        Map of dimensions to coordinates
    dims : dict of {str : list of str}, optional
        Map variable names to their coordinates. Will be inferred if they are not provided.
    pred_dims : dict, optional
        Dims for predictions data. Map variable names to their coordinates. Default behavior is to
        infer dims if this is not provided
    extra_event_dims : dict, optional
        Extra event dims for deterministic sites. Maps event dims that couldnt be inferred to
        their coordinates.
    num_chains : int, default 1
        Number of chains used for sampling. Ignored if posterior is present.

    Returns
    -------
    DataTree
    """
    posterior = SVIWrapper(
        svi,
        svi_result=svi_result,
        model_args=model_args,
        model_kwargs=model_kwargs,
        num_samples=num_samples,
    )
    with rc_context(rc={"data.sample_dims": ["sample"]}):
        return NumPyroConverter(
            posterior=posterior,
            prior=prior,
            posterior_predictive=posterior_predictive,
            predictions=predictions,
            constant_data=constant_data,
            predictions_constant_data=predictions_constant_data,
            log_likelihood=log_likelihood,
            index_origin=index_origin,
            coords=coords,
            dims=dims,
            pred_dims=pred_dims,
            extra_event_dims=extra_event_dims,
            num_chains=0,
        ).to_datatree()



# File generated with docstub

from typing import Any

import numpy as np
import xarray as xr
from _typeshed import Incomplete
from xarray import Dataset, DataTree, open_datatree

from arviz_base.base import dict_to_dataset

__all__ = [
    "convert_to_datatree",
    "convert_to_dataset",
]

def convert_to_datatree(obj: Incomplete, **kwargs: Incomplete) -> DataTree: ...
def convert_to_dataset(
    obj: Any, *, group: str = ..., **kwargs: Incomplete
) -> Dataset: ...



# File generated with docstub

import logging
import re
from collections.abc import Hashable, Mapping, Sequence
from pathlib import Path
from typing import Any

import cmdstanpy
import numpy as np
from _typeshed import Incomplete
from numpy.typing import ArrayLike
from xarray import DataTree

from arviz_base.base import dict_to_dataset, infer_stan_dtypes, requires
from arviz_base.rcparams import rcParams

_log: Incomplete

class CmdStanPyConverter:
    def __init__(
        self,
        *,
        posterior=...,
        posterior_predictive=...,
        predictions=...,
        prior=...,
        prior_predictive=...,
        observed_data=...,
        constant_data=...,
        predictions_constant_data=...,
        log_likelihood=...,
        index_origin=...,
        coords=...,
        dims=...,
        save_warmup=...,
        dtypes=...,
    ) -> None: ...
    def _warmup_return_to_dict(
        self, data: Incomplete, data_warmup: Incomplete, group: Incomplete
    ) -> None: ...
    def posterior_to_xarray(self) -> None: ...
    def sample_stats_to_xarray(self) -> None: ...
    def sample_stats_prior_to_xarray(self) -> None: ...
    def stats_to_xarray(self, fit) -> None: ...
    def posterior_predictive_to_xarray(self) -> None: ...
    def prior_predictive_to_xarray(self) -> None: ...
    def predictive_to_xarray(self, names, fit) -> None: ...
    def predictions_to_xarray(self) -> None: ...
    def log_likelihood_to_xarray(self) -> None: ...
    def prior_to_xarray(self) -> None: ...
    def observed_data_to_xarray(self) -> None: ...
    def constant_data_to_xarray(self) -> None: ...
    def predictions_constant_data_to_xarray(self) -> None: ...
    def to_datatree(self) -> None: ...

def _as_set(spec) -> None: ...
def _filter(names, spec) -> None: ...
def _unpack_fit(fit, items: list, save_warmup: bool, dtypes: dict) -> dict: ...
def from_cmdstanpy(
    posterior: cmdstanpy.CmdStanMCMC | None = ...,
    *,
    posterior_predictive: str | list[str] | None = ...,
    predictions: str | list[str] | None = ...,
    prior: cmdstanpy.CmdStanMCMC | None = ...,
    prior_predictive: str | list[str] | None = ...,
    observed_data: Mapping[str, ArrayLike] | None = ...,
    constant_data: Mapping[str, ArrayLike] | None = ...,
    predictions_constant_data: dict | None = ...,
    log_likelihood: str | list[str] | dict[str, str] | None = ...,
    index_origin: int | None = ...,
    coords: dict | None = ...,
    dims: Mapping[Any, Sequence[Hashable]] | None = ...,
    save_warmup: bool | None = ...,
    dtypes: dict | None = ...,
) -> DataTree: ...



# File generated with docstub

import warnings
from collections.abc import Hashable, Mapping, Sequence
from typing import Any

from numpy.typing import ArrayLike
from xarray import DataTree

from arviz_base.base import dict_to_dataset
from arviz_base.rcparams import rcParams

def from_dict(
    data: Mapping[Any, Mapping[Any, ArrayLike]],
    *,
    name: str | None = ...,
    sample_dims: Sequence[Hashable] | None = ...,
    save_warmup: bool | None = ...,
    index_origin: int | None = ...,
    coords: Mapping[Any, ArrayLike] | None = ...,
    dims: Mapping[Any, Sequence[Hashable]] | None = ...,
    pred_dims: Mapping[Any, Sequence[Hashable]] | None = ...,
    pred_coords: Mapping[Any, ArrayLike] | None = ...,
    check_conventions: bool = ...,
    attrs: Mapping[Any, Mapping[Any, Any]] | None = ...,
) -> DataTree: ...



# pylint: disable=unused-argument
"""Utilities to generate labels from xarray objects."""

from collections.abc import Hashable, Iterable, Mapping, Sequence
from typing import Any, Protocol

__all__ = [
    "mix_labellers",
    "Labeller",
    "BaseLabeller",
    "DimCoordLabeller",
    "IdxLabeller",
    "DimIdxLabeller",
    "MapLabeller",
    "NoVarLabeller",
]


class Labeller(Protocol):
    """Placeholder for type checking ``labeller`` used in docstrings."""

    def make_label_vert(  # noqa: D102
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...
    def make_label_flat(  # noqa: D102
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...


def mix_labellers(labellers, class_name="MixtureLabeller"):
    """Combine Labeller classes dynamically.

    The Labeller class aims to split plot labeling in ArviZ into atomic tasks to maximize
    extensibility, and the few classes provided are designed with small deviations
    from the base class, in many cases only one method is modified by the child class.
    It is to be expected then to want to use multiple classes "at once".

    This functions helps combine classes dynamically.

    For a general overview of ArviZ label customization, including
    ``mix_labellers``, see the :ref:`label_guide` page.

    Parameters
    ----------
    labellers : iterable of type
        Iterable of Labeller types to combine
    class_name : str, optional
        The name of the generated class

    Returns
    -------
    type
        Mixture class object. **It is not initialized**, and it should be
        initialized before passing it to ArviZ functions.

    Examples
    --------
    Combine the :class:`~arviz.labels.DimCoordLabeller` with the
    :class:`~arviz.labels.MapLabeller` to generate labels in the style of the
    ``DimCoordLabeller`` but using the mappings defined by ``MapLabeller``.
    Note that this works even though both modify the same methods because
    ``MapLabeller`` implements the mapping and then calls `super().method`.

    .. jupyter-execute::

        from arviz_base.labels import mix_labellers, DimCoordLabeller, MapLabeller
        l1 = DimCoordLabeller()
        sel = {"dim1": "a", "dim2": "top"}
        print(f"Output of DimCoordLabeller alone > {l1.sel_to_str(sel, sel)}")
        l2 = MapLabeller(dim_map={"dim1": "$d_1$", "dim2": r"$d_2$"})
        print(f"Output of MapLabeller alone > {l2.sel_to_str(sel, sel)}")
        l3 = mix_labellers(
            (MapLabeller, DimCoordLabeller)
        )(dim_map={"dim1": "$d_1$", "dim2": r"$d_2$"})
        print(f"Output of mixture labeller > {l3.sel_to_str(sel, sel)}")

    We can see how the mappings are taken into account as well as the dim+coord style.
    However, the order in the ``labellers`` arg iterator is important!
    See for yourself:

    .. jupyter-execute::

        l4 = mix_labellers(
            (DimCoordLabeller, MapLabeller)
        )(dim_map={"dim1": "$d_1$", "dim2": r"$d_2$"})
        print(f"Output of inverted mixture labeller > {l4.sel_to_str(sel, sel)}")

    """
    return type(class_name, labellers, {})


class BaseLabeller:
    """Base labeller class.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" are:

    .. code-block:: none
        :caption: Single line label

        theta[0, Name]

    .. code-block:: none
        :caption: Multi-line label

        theta
        0, Name

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def dim_coord_to_str(self, dim, coord_val, coord_idx):  # pylint: disable=no-self-use
        """Format a single dimension name, its value and positional indexes as a string.

        Parameters
        ----------
        dim : hashable
            Dimension name
        coord_val : hashable
            Coordinate label
        coord_idx : int or sequence of int
            Positions along `dim` where its coordinate is `coord_val`

        Returns
        -------
        str
        """
        return f"{coord_val}"

    def sel_to_str(self, sel, isel):
        """Format selection dictionaries as a string.

        Parameters
        ----------
        sel : mapping of {hashable_key : hashable}
        isel : mapping of {hashable_key : int or sequence of int}

        Returns
        -------
        str
        """
        if sel:
            return ", ".join(
                [
                    self.dim_coord_to_str(dim, v, i)
                    for (dim, v), (_, i) in zip(sel.items(), isel.items())
                ]
            )
        return ""

    def var_name_to_str(self, var_name):  # pylint: disable=no-self-use
        """Format a variable name as a string.

        Parameters
        ----------
        var_name : str or None
            The variable name. It should accept ``None``

        Returns
        -------
        str or None
        """
        return var_name

    def var_pp_to_str(self, var_name, pp_var_name):
        """Format the corresponding variable names for observation and posterior predictive.

        Parameters
        ----------
        var_name : str or None
        pp_var_name : str or None

        Returns
        -------
        str or None
        """
        var_name_str = self.var_name_to_str(var_name)
        pp_var_name_str = self.var_name_to_str(pp_var_name)
        if var_name_str is None and pp_var_name_str is None:
            return None
        if var_name_str == pp_var_name_str:
            return var_name_str
        return f"{var_name_str} / {pp_var_name_str}"

    def make_label_vert(self, var_name, sel, isel):
        """Format variable name and corresponding subset as a multiline string.

        Parameters
        ----------
        var_name : str or None
        sel : mapping of {hashable_key : hashable}
        isel : mapping of {hashable_key : int or sequence of int}

        Returns
        -------
        str
        """
        var_name_str = self.var_name_to_str(var_name)
        sel_str = self.sel_to_str(sel, isel)
        if not sel_str:
            return "" if var_name_str is None else var_name_str
        if var_name_str is None:
            return sel_str
        return f"{var_name_str}\n{sel_str}"

    def make_label_flat(self, var_name, sel, isel):
        """Format variable name and corresponding subset as a single line string.

        Parameters
        ----------
        var_name : str or None
        sel : mapping of {hashable_key : hashable}
        isel : mapping of {hashable_key : int or sequence of int}

        Returns
        -------
        str
        """
        var_name_str = self.var_name_to_str(var_name)
        sel_str = self.sel_to_str(sel, isel)
        if not sel_str:
            return "" if var_name_str is None else var_name_str
        if var_name_str is None:
            return sel_str
        return f"{var_name_str}[{sel_str}]"

    def make_pp_label(self, var_name, pp_var_name, sel, isel):
        """Format obs+pp variable name plus corresponding subsets as a multiline string.

        Parameters
        ----------
        var_name, pp_var_name : str or None
        sel : mapping of {hashable_key : hashable}
        isel : mapping of {hashable_key : int or sequence of int}

        Returns
        -------
        str
        """
        names = self.var_pp_to_str(var_name, pp_var_name)
        return self.make_label_vert(names, sel, isel)


class DimCoordLabeller(BaseLabeller):
    """Labeller class to show both dimension and coordinate value information.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" are:

    .. code-block:: none
        :caption: Single line label

        theta[chain: 0, school: Name]

    .. code-block:: none
        :caption: Multi-line label

        theta
        chain: 0, school: Name

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def dim_coord_to_str(self, dim, coord_val, coord_idx):
        """Format a single dimension name, its value and positional indexes as a string.

        Parameters
        ----------
        dim : hashable
            Dimension name
        coord_val : hashable
            Coordinate label
        coord_idx : int or sequence of int
            Positions along `dim` where its coordinate is `coord_val`

        Returns
        -------
        str
        """
        return f"{dim}: {coord_val}"


class IdxLabeller(BaseLabeller):
    """Labeller class to show positional index information.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" (3rd and 5th element in the school coordinate values) are:

    .. code-block:: none
        :caption: Single line label

        theta[0, 2,4]

    .. code-block:: none
        :caption: Multi-line label

        theta
        0, 2,4

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def dim_coord_to_str(self, dim, coord_val, coord_idx):
        """Format a single dimension name, its value and positional indexes as a string.

        Parameters
        ----------
        dim : hashable
            Dimension name
        coord_val : hashable
            Coordinate label
        coord_idx : int or sequence of int
            Positions along `dim` where its coordinate is `coord_val`

        Returns
        -------
        str
        """
        if not isinstance(coord_idx, Iterable):
            return f"{coord_idx}"
        return f"{','.join(str(idx) for idx in coord_idx)}"


class DimIdxLabeller(BaseLabeller):
    """Labeller class to show both dimension and positional index information.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" (3rd and 5th element in the school coordinate values) are:

    .. code-block:: none
        :caption: Single line label

        theta[chain#0, school#2,4]

    .. code-block:: none
        :caption: Multi-line label

        theta
        chain#0, school#2,4

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def dim_coord_to_str(self, dim, coord_val, coord_idx):
        """Format a single dimension name, its value and positional indexes as a string.

        Parameters
        ----------
        dim : hashable
            Dimension name
        coord_val : hashable
            Coordinate label
        coord_idx : int or sequence of int
            Positions along `dim` where its coordinate is `coord_val`

        Returns
        -------
        str
        """
        if not isinstance(coord_idx, Iterable):
            return f"{dim}#{coord_idx}"
        return f"{dim}#{','.join(str(idx) for idx in coord_idx)}"


class MapLabeller(BaseLabeller):
    """Labeller class to perform provided replacements to elements when converting to string.

    It is a subclass of :class:`BaseLabeller` so the base behaviour is the same,
    but we can define replacements through dictionaries when initializing the class.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" with a replacement mapping on the variable name ``{"theta": "θ"}``
    and one on the coordinate values ``{"Name": "𝑁𝑎𝑚𝑒"}``

    .. code-block:: none
        :caption: Single line label

        θ[0, 𝑁𝑎𝑚𝑒]

    .. code-block:: none
        :caption: Multi-line label

        θ
        0, 𝑁𝑎𝑚𝑒

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def __init__(self, var_name_map=None, dim_map=None, coord_map=None):
        """Initialize a MapLabeller class.

        Parameters
        ----------
        var_name_map, dim_map : mapping of {hashable : hashable}, optional
            Keys are existing names and values are their respective desired labels.
        coord_map : mapping of {hashable : mapping of {hashable : hashable}}, optional
            The keys of the first level dictionary are dimension names, the inner
            dictionary has existing coord names as keys and their corresponding desired
            label as values.
        """
        self.var_name_map = {} if var_name_map is None else var_name_map
        self.dim_map = {} if dim_map is None else dim_map
        self.coord_map = {} if coord_map is None else coord_map

    def dim_coord_to_str(self, dim, coord_val, coord_idx):
        """Format a single dimension name, its value and positional indexes as a string.

        Parameters
        ----------
        dim : hashable
            Dimension name
        coord_val : hashable
            Coordinate label
        coord_idx : int or sequence of int
            Positions along `dim` where its coordinate is `coord_val`

        Returns
        -------
        str
        """
        dim_str = self.dim_map.get(dim, dim)
        coord_str = self.coord_map.get(dim, {}).get(coord_val, coord_val)
        return super().dim_coord_to_str(dim_str, coord_str, coord_idx)

    def var_name_to_str(self, var_name):
        """Format a variable name as a string.

        Parameters
        ----------
        var_name : str or None
            The variable name. It should accept ``None``

        Returns
        -------
        str or None
        """
        var_name_str = self.var_name_map.get(var_name, var_name)
        return super().var_name_to_str(var_name_str)


class NoVarLabeller(BaseLabeller):
    """Labeller class to exclude the variable name from the generated labels.

    The default labels for "theta" variable on the subset corresponding to "chain" 0
    and "school" "Name" are:

    .. code-block:: none
        :caption: Single line label

        0, Name

    .. code-block:: none
        :caption: Multi-line label

        0, Name

    See Also
    --------
    :ref:`label_guide`: Tutorial page on using labellers with ArviZ
    """

    def var_name_to_str(self, var_name):
        """Format a variable name as a string.

        Parameters
        ----------
        var_name : str or None
            The variable name. It should accept ``None``

        Returns
        -------
        str or None
        """
        return None



"""Transform helper."""

from collections.abc import Callable
from copy import deepcopy

import xarray as xr

from arviz_base.utils import _var_names


def get_unconstrained_samples(
    idata,
    var_names=None,
    filter_vars=None,
    transform_funcs: dict[str, Callable[[xr.DataArray], xr.DataArray]] | None = None,
    group="posterior",
    return_dataset=False,
):
    """Transform helper.

    It transform the variables of a DataTree in a given group applying the functions
    in the dictionary `transform_funcs`.
    With the default values, it would combine the samples in posterior and unconstrained posterior
    to have all the samples in the unconstrained space.

    Parameters
    ----------
    idata : DataTree-like
        DataTree from which to extract the data.
    var_names : str or list of str, optional
        Variables to be extracted. Prefix the variables by `~` when you want to exclude them.
    filter_vars : {None, "like", "regex"}, optional
        If `None` (default), interpret var_names as the real variables names. If "like",
        interpret var_names as substrings of the real variables names. If "regex",
        interpret var_names as regular expressions on the real variables names. A la
        `pandas.filter`.
    transform_funcs : dict of {str : Callable}, optional
        Dictionary with the functions to transform each variable, by default None.
        Each function *must* accept a single DataArray and return also a
        DataArray.
        If `None`, the values are taking from the group `unconstrained_{group}.
    group : str, optional
        The group where the transformation will be applied, by default `posterior`
    return_dataset : bool, optional
        If true it will return a Dataset with the transformed samples, by default False.

    Returns
    -------
    DataTree-like or Dataset
        By defaults, it returns the original DataTree with the transformed data in
        the `unconstrained_{group}` group.
        A Dataset with the transformed samples.
    """
    if transform_funcs is None:
        transform_funcs = {}

    # get the two datasets and extract variables
    idata = deepcopy(idata)
    ds_in = idata[group].to_dataset()
    var_names_in = _var_names(var_names, ds_in, filter_vars)
    if var_names_in:
        ds_in = ds_in[var_names_in]
    ds_out = idata.get(f"unconstrained_{group}", None)
    if ds_out:
        ds_out = ds_out.to_dataset()

    # for each variable in the input group, decide what to store
    new_vars = {}
    for name, da in ds_in.data_vars.items():
        if name in transform_funcs:
            # validate function
            func = transform_funcs[name]
            if not callable(func):
                raise TypeError(f"transform_funcs[{name!r}] must be callable, got {type(func)}")
            # apply user‐supplied transform to the datarray
            new_da = func(da)
            if not isinstance(new_da, xr.DataArray):
                raise TypeError(
                    f"transform_funcs[{name!r}] must return an xarray.DataArray,"
                    f" got {type(new_da)} instead."
                )
        elif ds_out is not None and name in ds_out.data_vars:
            # reuse the existing unconstrained version
            new_da = ds_out[name]
        else:
            # no transform, no unconstrained: copy the original
            new_da = da

        new_vars[name] = new_da

    # rebuild a Dataset for the return_group (keep the same coords)
    ds_new = xr.Dataset(new_vars, attrs=ds_in.attrs)

    if return_dataset:
        # return dataset
        return ds_new
    else:
        # return datatree in the unconstrained_{group}
        idata[f"unconstrained_{group}"] = xr.DataTree(ds_new)
        return idata



# File generated with docstub

import logging

from _typeshed import Incomplete

_log: Incomplete

from arviz_base import labels, testing
from arviz_base._version import __version__
from arviz_base.base import (
    dict_to_dataset,
    generate_dims_coords,
    make_attrs,
    ndarray_to_dataarray,
)
from arviz_base.citations import citations
from arviz_base.converters import convert_to_dataset, convert_to_datatree
from arviz_base.datasets import (
    clear_data_home,
    get_data_home,
    list_datasets,
    load_arviz_data,
)
from arviz_base.io_cmdstanpy import from_cmdstanpy
from arviz_base.io_dict import from_dict
from arviz_base.io_emcee import from_emcee
from arviz_base.io_numpyro import from_numpyro, from_numpyro_svi
from arviz_base.io_pystan import from_pystan
from arviz_base.rcparams import rc_context, rcParams
from arviz_base.reorg import (
    dataset_to_dataarray,
    dataset_to_dataframe,
    explode_dataset_dims,
    extract,
    references_to_dataset,
)
from arviz_base.sel_utils import xarray_sel_iter, xarray_to_ndarray, xarray_var_iter
from arviz_base.transform import get_unconstrained_samples

__all__ = [
    "__version__",
    "citations",
    "dict_to_dataset",
    "generate_dims_coords",
    "make_attrs",
    "ndarray_to_dataarray",
    "convert_to_dataset",
    "convert_to_datatree",
    "clear_data_home",
    "get_data_home",
    "list_datasets",
    "load_arviz_data",
    "from_cmdstanpy",
    "from_dict",
    "from_emcee",
    "from_numpyro",
    "from_numpyro_svi",
    "labels",
    "rc_context",
    "rcParams",
    "extract",
    "dataset_to_dataarray",
    "dataset_to_dataframe",
    "explode_dataset_dims",
    "references_to_dataset",
    "xarray_sel_iter",
    "xarray_var_iter",
    "xarray_to_ndarray",
    "testing",
    "get_unconstrained_samples",
]



"""Generalistic converters.

Here "generalistic" means catch anything that can be converter into datatree and
convert it via its specific function.
"""

import numpy as np
import xarray as xr
from xarray import DataTree, open_datatree

from arviz_base.base import dict_to_dataset

__all__ = [
    "convert_to_datatree",
    "convert_to_dataset",
]


# pylint: disable=too-many-return-statements
def convert_to_datatree(obj, **kwargs):
    r"""Convert a supported object to a DataTree object following ArviZ conventions.

    This function sends `obj` to the right conversion function. It is idempotent,
    in that it will return DataTree objects unchanged. In general however,
    it is better to call specific conversion functions directly. See below
    for more details.

    Parameters
    ----------
    obj
        A supported object to convert to InferenceData:

         * DataTree: returns unchanged
         * InferenceData: returns the equivalent DataTree. `kwargs` are passed
           to :meth:`datatree.DataTree.from_dict`.
         * str:

           - If it ends with ``.csv``, attempts to load the file as a cmdstan csv fit
             using :func:`from_cmdstan`
           - Otherwise, attempts to load a netcdf or zarr file from disk
             using :func:`open_datatree`

         * pystan fit: Calls :func:`.from_pystan` with default arguments
         * cmdstanpy fit: Calls :func:`from_cmdstanpy` with default arguments
         * cmdstan csv-list: Calls :func:`from_cmdstan` with default arguments
         * emcee sampler: Calls :func:`from_emcee` with default arguments
         * pyro MCMC: Calls :func:`from_pyro` with default arguments
         * numpyro MCMC: calls :func:`from_numpyro` with default arguments
         * beanmachine MonteCarloSamples: Calls :func:`from_beanmachine` with default arguments
         * `xarray.Dataset`: Adds it to the DataTree a the only group. The group name
           is taken from the ``group`` keyword in `kwargs`.
         * `xarray.DataArray`: Adds it to the DataTree as the only variable in a single group.
           If the ``name`` is not set, "x" is used as name. Like above,
           the group name is taken from the ``group`` keyword in `kwargs`.
         * dict: creates an xarray.Dataset with :func:`dict_to_dataset` and adds it
           to the DataTree as the only group (named with the ``group`` key in `kwargs`).
         * `numpy.ndarray`: names the variable "x" and adds it to the DataTree
           with a single group, named with the ``group`` key in `kwargs`.

    **kwargs
        Rest of the supported keyword arguments transferred to conversion function.

    Returns
    -------
    DataTree

    See Also
    --------
    from_dict
        Convert a nested dictionary of {group_name: {var_name: data}} to a DataTree.
    """
    kwargs = kwargs.copy()
    group = kwargs.pop("group", "posterior")

    # Cases that convert to DataTree
    if isinstance(obj, DataTree):
        return obj
    if isinstance(obj, str):
        # if obj.endswith(".csv"):
        #     if group == "sample_stats":
        #         kwargs["posterior"] = obj
        #     elif group == "sample_stats_prior":
        #         kwargs["prior"] = obj
        #     return from_cmdstan(**kwargs)
        return open_datatree(obj, **kwargs)
    if obj.__class__.__name__ == "InferenceData":
        return DataTree.from_dict({group: obj[group] for group in obj.groups()}, **kwargs)
    # if (
    #     obj.__class__.__name__ in {"StanFit4Model", "CmdStanMCMC"}
    #     or obj.__class__.__module__ == "stan.fit"
    # ):
    #     if group == "sample_stats":
    #         kwargs["posterior"] = obj
    #     elif group == "sample_stats_prior":
    #         kwargs["prior"] = obj
    #     if obj.__class__.__name__ == "CmdStanMCMC":
    #         return from_cmdstanpy(**kwargs)
    #     return from_pystan(**kwargs)
    # if obj.__class__.__name__ == "EnsembleSampler":  # ugly, but doesn't make emcee a requirement
    #     return from_emcee(sampler=obj, **kwargs)
    # if obj.__class__.__name__ == "MonteCarloSamples":
    #     return from_beanmachine(sampler=obj, **kwargs)
    # if obj.__class__.__name__ == "MCMC" and obj.__class__.__module__.startswith("pyro"):
    #     return from_pyro(posterior=obj, **kwargs)
    # if obj.__class__.__name__ == "MCMC" and obj.__class__.__module__.startswith("numpyro"):
    #     return from_numpyro(posterior=obj, **kwargs)

    # Cases that convert to xarray
    if isinstance(obj, xr.Dataset):
        dataset = obj
    elif isinstance(obj, xr.DataArray):
        if obj.name is None:
            obj.name = "x"
        dataset = obj.to_dataset()
    elif isinstance(obj, dict):
        dataset = dict_to_dataset(obj, **kwargs)
    elif isinstance(obj, np.ndarray):
        dataset = dict_to_dataset({"x": obj}, **kwargs)
    # elif isinstance(obj, (list, tuple)) and isinstance(obj[0], str) and obj[0].endswith(".csv"):
    #     if group == "sample_stats":
    #         kwargs["posterior"] = obj
    #     elif group == "sample_stats_prior":
    #         kwargs["prior"] = obj
    #     return from_cmdstan(**kwargs)
    else:
        allowable_types = (
            "xarray dataarray",
            "xarray dataset",
            "dict",
            "netcdf filename",
            "zarr filename",
            "numpy array",
            # "pystan fit",
            # "emcee fit",
            # "pyro mcmc fit",
            # "numpyro mcmc fit",
            # "cmdstan fit csv filename",
            # "cmdstanpy fit",
            # "beanmachine montecarlosamples",
        )
        raise ValueError(
            f"Can only convert {', '.join(allowable_types)} to InferenceData, "
            f"not {obj.__class__.__name__}"
        )

    return DataTree.from_dict({group: dataset})


def convert_to_dataset(obj, *, group="posterior", **kwargs):
    """Convert a supported object to an xarray.Dataset.

    This function is idempotent: if you pass in an xarray.Dataset, it returns it unchanged.
    If passed another supported object (e.g., dict, DataTree, numpy array), it will be
    converted to a Dataset via `convert_to_datatree`, and the specified group will be extracted.

    Parameters
    ----------
    obj : Any
        A supported object to convert to InferenceData. See `convert_to_datatree` for full list.
    group : str, default "posterior"
        The group name to extract from the converted DataTree.
    **kwargs
        Additional arguments passed to `convert_to_datatree`.

    Returns
    -------
    xarray.Dataset
        The extracted dataset from the converted object.

    Raises
    ------
    ValueError
        If the group cannot be extracted from the resulting DataTree.

    Examples
    --------
    >>> convert_to_dataset({"mu": np.random.randn(500)})
    <xarray.Dataset> ...  # Posterior group with 'mu' variable
    """
    if isinstance(obj, DataTree) and obj.name == group:
        return obj.to_dataset()
    inference_data = convert_to_datatree(obj, group=group, **kwargs)
    dataset = getattr(inference_data, group, None)
    if dataset is None:
        raise ValueError(
            f"Can not extract {group} from {obj}! See docs for other conversion utilities."
        )
    return dataset.to_dataset()



# File generated with docstub

import numpy as np
from xarray import DataTree

from arviz_base import from_dict

def generate_base_data(seed=...) -> None: ...
def datatree(seed=...) -> None: ...
def datatree2(seed=...) -> None: ...
def datatree3(seed=...) -> None: ...
def datatree_binary(seed=...) -> None: ...
def datatree_regression(seed=...) -> None: ...
def datatree_4d(seed=...) -> None: ...
def datatree_sample(seed=...) -> None: ...
def cmp() -> None: ...
def fake_dt() -> None: ...
def datatree_censored(seed=...) -> None: ...
def check_multiple_attrs(
    test_dict: dict[str, list[str]], parent: DataTree
) -> list[str | tuple[str, str]]: ...



#  pylint: disable=too-many-instance-attributes,too-many-lines
"""PyStan-specific conversion code."""

import logging
import re
from math import ceil

import numpy as np
from xarray import DataArray, Dataset, DataTree

from arviz_base.base import (
    dict_to_dataset,
    generate_dims_coords,
    infer_stan_dtypes,
    make_attrs,
    requires,
)
from arviz_base.rcparams import rcParams

_log = logging.getLogger(__name__)


class PyStanConverter:
    """Encapsulate PyStan specific logic."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        *,
        posterior=None,
        posterior_model=None,
        posterior_predictive=None,
        predictions=None,
        prior=None,
        prior_model=None,
        prior_predictive=None,
        observed_data=None,
        constant_data=None,
        predictions_constant_data=None,
        log_likelihood=False,
        coords=None,
        dims=None,
        save_warmup=None,
        dtypes=None,
    ):
        self.posterior = posterior
        self.posterior_model = posterior_model
        self.posterior_predictive = posterior_predictive
        self.predictions = predictions
        self.prior = prior
        self.prior_model = prior_model
        self.prior_predictive = prior_predictive
        self.observed_data = observed_data
        self.constant_data = constant_data
        self.predictions_constant_data = predictions_constant_data
        self.log_likelihood = log_likelihood
        self.coords = coords
        self.dims = dims
        self.save_warmup = rcParams["data.save_warmup"] if save_warmup is None else save_warmup
        self.dtypes = dtypes

        if (
            self.log_likelihood is True
            and self.posterior is not None
            and "log_lik" in self.posterior.param_names
        ):
            self.log_likelihood = ["log_lik"]
        elif isinstance(self.log_likelihood, bool):
            self.log_likelihood = None

        import stan  # pylint: disable=import-error

        self.stan = stan

    def _warmup_return_to_dict(self, data, data_warmup, group, attrs=None):
        res = {
            group: dict_to_dataset(
                data,
                inference_library=self.stan,
                coords=self.coords,
                dims=self.dims,
                skip_event_dims="log_likelihood" == group,
                attrs=attrs,
            ),
        }
        if self.save_warmup and data_warmup:
            res[f"warmup_{group}"] = dict_to_dataset(
                data_warmup,
                inference_library=self.stan,
                coords=self.coords,
                dims=self.dims,
                skip_event_dims="log_likelihood" == group,
                attrs=attrs,
            )
        return res

    @requires("posterior")
    def posterior_to_xarray(self):
        """Extract posterior samples from fit."""
        posterior = self.posterior
        posterior_model = self.posterior_model
        # filter posterior_predictive and log_likelihood
        posterior_predictive = self.posterior_predictive
        if posterior_predictive is None:
            posterior_predictive = []
        elif isinstance(posterior_predictive, str):
            posterior_predictive = [posterior_predictive]
        predictions = self.predictions
        if predictions is None:
            predictions = []
        elif isinstance(predictions, str):
            predictions = [predictions]
        log_likelihood = self.log_likelihood
        if log_likelihood is None:
            log_likelihood = []
        elif isinstance(log_likelihood, str):
            log_likelihood = [log_likelihood]
        elif isinstance(log_likelihood, dict):
            log_likelihood = list(log_likelihood.values())

        ignore = posterior_predictive + predictions + log_likelihood

        data, data_warmup = get_draws(
            posterior,
            model=posterior_model,
            ignore=ignore,
            warmup=self.save_warmup,
            dtypes=self.dtypes,
        )
        attrs = get_attrs(posterior, model=posterior_model)
        return self._warmup_return_to_dict(data, data_warmup, "posterior", attrs)

    @requires("posterior")
    def sample_stats_to_xarray(self):
        """Extract sample_stats from posterior."""
        posterior = self.posterior
        posterior_model = self.posterior_model
        data, data_warmup = get_sample_stats(
            posterior, ignore="lp__", warmup=self.save_warmup, dtypes=self.dtypes
        )
        data_lp, data_warmup_lp = get_sample_stats(
            posterior, variables="lp__", warmup=self.save_warmup
        )
        data["lp"] = data_lp["lp"]
        if data_warmup_lp:
            data_warmup["lp"] = data_warmup_lp["lp"]

        attrs = get_attrs(posterior, model=posterior_model)
        return self._warmup_return_to_dict(data, data_warmup, "sample_stats", attrs)

    @requires("posterior")
    @requires("log_likelihood")
    def log_likelihood_to_xarray(self):
        """Store log_likelihood data in log_likelihood group."""
        fit = self.posterior

        log_likelihood = self.log_likelihood
        model = self.posterior_model
        if isinstance(log_likelihood, str):
            log_likelihood = [log_likelihood]
        if isinstance(log_likelihood, (list | tuple)):
            log_likelihood = {name: name for name in log_likelihood}
        log_likelihood_draws, log_likelihood_draws_warmup = get_draws(
            fit,
            model=model,
            variables=list(log_likelihood.values()),
            warmup=self.save_warmup,
            dtypes=self.dtypes,
        )
        data = {
            obs_var_name: log_likelihood_draws[log_like_name]
            for obs_var_name, log_like_name in log_likelihood.items()
            if log_like_name in log_likelihood_draws
        }
        data_warmup = {
            obs_var_name: log_likelihood_draws_warmup[log_like_name]
            for obs_var_name, log_like_name in log_likelihood.items()
            if log_like_name in log_likelihood_draws_warmup
        }
        return self._warmup_return_to_dict(data, data_warmup, "log_likelihood")

    @requires("posterior")
    @requires("posterior_predictive")
    def posterior_predictive_to_xarray(self):
        """Convert posterior_predictive samples to xarray."""
        posterior = self.posterior
        posterior_model = self.posterior_model
        posterior_predictive = self.posterior_predictive
        data, data_warmup = get_draws(
            posterior,
            model=posterior_model,
            variables=posterior_predictive,
            warmup=self.save_warmup,
            dtypes=self.dtypes,
        )
        return self._warmup_return_to_dict(data, data_warmup, "posterior_predictive")

    @requires("posterior")
    @requires("predictions")
    def predictions_to_xarray(self):
        """Convert predictions samples to xarray."""
        posterior = self.posterior
        posterior_model = self.posterior_model
        predictions = self.predictions
        data, data_warmup = get_draws(
            posterior,
            model=posterior_model,
            variables=predictions,
            warmup=self.save_warmup,
            dtypes=self.dtypes,
        )
        return self._warmup_return_to_dict(data, data_warmup, "predictions")

    @requires("prior")
    def prior_to_xarray(self):
        """Convert prior samples to xarray."""
        prior = self.prior
        prior_model = self.prior_model
        # filter posterior_predictive and log_likelihood
        prior_predictive = self.prior_predictive
        if prior_predictive is None:
            prior_predictive = []
        elif isinstance(prior_predictive, str):
            prior_predictive = [prior_predictive]

        ignore = prior_predictive

        data, data_warmup = get_draws(
            prior, model=prior_model, ignore=ignore, warmup=self.save_warmup, dtypes=self.dtypes
        )
        attrs = get_attrs(prior, model=prior_model)
        return self._warmup_return_to_dict(data, data_warmup, "prior", attrs)

    @requires("prior")
    def sample_stats_prior_to_xarray(self):
        """Extract sample_stats_prior from prior."""
        prior = self.prior
        prior_model = self.prior_model
        data, data_warmup = get_sample_stats(prior, warmup=self.save_warmup, dtypes=self.dtypes)
        attrs = get_attrs(prior, model=prior_model)
        return self._warmup_return_to_dict(data, data_warmup, "sample_stats_prior", attrs)

    @requires("prior")
    @requires("prior_predictive")
    def prior_predictive_to_xarray(self):
        """Convert prior_predictive samples to xarray."""
        prior = self.prior
        prior_model = self.prior_model
        prior_predictive = self.prior_predictive
        data, data_warmup = get_draws(
            prior,
            model=prior_model,
            variables=prior_predictive,
            warmup=self.save_warmup,
            dtypes=self.dtypes,
        )
        return self._warmup_return_to_dict(data, data_warmup, "prior_predictive")

    @requires("posterior_model")
    @requires(["observed_data"])
    def observed_data_to_xarray(self):
        """Convert observed data to xarray."""
        posterior_model = self.posterior_model
        dims = {} if self.dims is None else self.dims
        names = getattr(self, "observed_data")
        if names is None:
            return None
        names = [names] if isinstance(names, str) else names
        data = {}
        for key in names:
            vals = np.atleast_1d(posterior_model.data[key])
            val_dims = dims.get(key)
            val_dims, coords = generate_dims_coords(
                vals.shape, key, dims=val_dims, coords=self.coords
            )
            data[key] = DataArray(vals, dims=val_dims, coords=coords)
        return Dataset(data_vars=data, attrs=make_attrs(inference_library=self.stan))

    @requires("posterior_model")
    @requires(["constant_data"])
    def constant_data_to_xarray(self):
        """Convert observed data to xarray."""
        posterior_model = self.posterior_model
        dims = {} if self.dims is None else self.dims
        names = getattr(self, "constant_data")
        if names is None:
            return None
        names = [names] if isinstance(names, str) else names
        data = {}
        for key in names:
            vals = np.atleast_1d(posterior_model.data[key])
            val_dims = dims.get(key)
            val_dims, coords = generate_dims_coords(
                vals.shape, key, dims=val_dims, coords=self.coords
            )
            data[key] = DataArray(vals, dims=val_dims, coords=coords)
        return Dataset(data_vars=data, attrs=make_attrs(inference_library=self.stan))

    @requires("posterior_model")
    @requires("predictions_constant_data")
    def predictions_constant_data_to_xarray(self):
        """Convert observed data to xarray."""
        posterior_model = self.posterior_model
        dims = {} if self.dims is None else self.dims
        names = self.predictions_constant_data
        names = [names] if isinstance(names, str) else names
        data = {}
        for key in names:
            vals = np.atleast_1d(posterior_model.data[key])
            val_dims = dims.get(key)
            val_dims, coords = generate_dims_coords(
                vals.shape, key, dims=val_dims, coords=self.coords
            )
            data[key] = DataArray(vals, dims=val_dims, coords=coords)
        return Dataset(data_vars=data, attrs=make_attrs(inference_library=self.stan))

    def to_datatree(self):
        """Convert all available data to a DataTree object.

        Note that if groups can not be created (i.e., there is no `output`, so
        the `posterior` and `sample_stats` can not be extracted), then the DataTree
        will not have those groups.
        """
        datadict = {
            "observed_data": self.observed_data_to_xarray(),
            "constant_data": self.constant_data_to_xarray(),
            "predictions_constant_data": self.predictions_constant_data_to_xarray(),
        }
        datalist = [
            self.posterior_to_xarray(),
            self.sample_stats_to_xarray(),
            self.posterior_predictive_to_xarray(),
            self.predictions_to_xarray(),
            self.prior_to_xarray(),
            self.sample_stats_prior_to_xarray(),
            self.prior_predictive_to_xarray(),
            self.log_likelihood_to_xarray(),
        ]
        for ds_dict in datalist:
            if ds_dict is not None:
                datadict.update(ds_dict)
        return DataTree.from_dict({group: ds for group, ds in datadict.items() if ds is not None})


def get_draws(fit, model=None, variables=None, ignore=None, warmup=False, dtypes=None):
    """Extract draws from PyStan fit."""
    if ignore is None:
        ignore = []

    if dtypes is None:
        dtypes = {}

    if model is not None:
        dtypes = {**infer_dtypes(fit, model), **dtypes}

    if not fit.save_warmup:
        warmup = False

    num_warmup = ceil((fit.num_warmup * fit.save_warmup) / fit.num_thin)

    if variables is None:
        variables = fit.param_names
    elif isinstance(variables, str):
        variables = [variables]
    variables = list(variables)

    data = {}
    data_warmup = {}

    for var in variables:
        if var in ignore:
            continue
        if var in data:
            continue
        dtype = dtypes.get(var)

        new_shape = (*fit.dims[fit.param_names.index(var)], -1, fit.num_chains)
        if 0 in new_shape:
            continue
        values = fit._draws[fit._parameter_indexes(var), :]  # pylint: disable=protected-access
        values = values.reshape(new_shape, order="F")
        values = np.moveaxis(values, [-2, -1], [1, 0])
        values = values.astype(dtype)
        if warmup:
            data_warmup[var] = values[:, num_warmup:]
        data[var] = values[:, num_warmup:]

    return data, data_warmup


def get_sample_stats(fit, variables=None, ignore=None, warmup=False, dtypes=None):
    """Extract sample stats from PyStan fit."""
    if dtypes is None:
        dtypes = {}
    dtypes = {"divergent__": bool, "n_leapfrog__": np.int64, "treedepth__": np.int64, **dtypes}

    rename_dict = {
        "divergent": "diverging",
        "n_leapfrog": "n_steps",
        "treedepth": "tree_depth",
        "stepsize": "step_size",
        "accept_stat": "acceptance_rate",
    }

    if isinstance(variables, str):
        variables = [variables]
    if isinstance(ignore, str):
        ignore = [ignore]

    if not fit.save_warmup:
        warmup = False

    num_warmup = ceil((fit.num_warmup * fit.save_warmup) / fit.num_thin)

    data = {}
    data_warmup = {}
    for key in fit.sample_and_sampler_param_names:
        if (variables and key not in variables) or (ignore and key in ignore):
            continue
        new_shape = -1, fit.num_chains
        values = fit._draws[fit._parameter_indexes(key)]  # pylint: disable=protected-access
        values = values.reshape(new_shape, order="F")
        values = np.moveaxis(values, [-2, -1], [1, 0])
        dtype = dtypes.get(key)
        values = values.astype(dtype)
        name = re.sub("__$", "", key)
        name = rename_dict.get(name, name)
        if warmup:
            data_warmup[name] = values[:, :num_warmup]
        data[name] = values[:, num_warmup:]

    return data, data_warmup


def get_attrs(fit, model=None):
    """Get attributes from PyStan fit and model object."""
    attrs = {}
    for key in ["num_chains", "num_samples", "num_thin", "num_warmup", "save_warmup"]:
        try:
            attrs[key] = getattr(fit, key)
        except AttributeError as exp:
            _log.warning("Failed to access attribute %s in fit object %s", key, exp)

    if model is not None:
        for key in ["model_name", "program_code", "random_seed"]:
            try:
                attrs[key] = getattr(model, key)
            except AttributeError as exp:
                _log.warning("Failed to access attribute %s in model object %s", key, exp)

    return attrs


def infer_dtypes(fit, model=None):
    """Infer dtypes from Stan model code.

    Function strips out generated quantities block and searches for `int`
    dtypes after stripping out comments inside the block.
    """
    if model is None:
        return {}
    stan_code = model.program_code
    model_pars = fit.param_names

    dtypes = {key: item for key, item in infer_stan_dtypes(stan_code).items() if key in model_pars}
    return dtypes


# pylint disable=too-many-instance-attributes
def from_pystan(
    posterior=None,
    *,
    posterior_predictive=None,
    predictions=None,
    prior=None,
    prior_predictive=None,
    observed_data=None,
    constant_data=None,
    predictions_constant_data=None,
    log_likelihood=None,
    coords=None,
    dims=None,
    posterior_model=None,
    prior_model=None,
    save_warmup=None,
    dtypes=None,
):
    """Convert PyStan data into an InferenceData object.

    For a usage example read the
    :ref:`Creating InferenceData section on from_pystan <creating_InferenceData>`

    Parameters
    ----------
    posterior : stan.fit.Fit
        PyStan fit object for posterior.
    posterior_predictive : str, a list of str
        Posterior predictive samples for the posterior.
    predictions : str, a list of str
        Out-of-sample predictions for the posterior.
    prior : stan.fit.Fit
        PyStan fit object for prior.
    prior_predictive : str, a list of str
        Posterior predictive samples for the prior.
    observed_data : str, a list of str
        observed data used in the sampling.
        Observed data is extracted from the `posterior.data`.
        PyStan needs model object for the extraction.
        See `posterior_model`.
    constant_data : str, a list of str
        Constants relevant to the model (i.e. x values in a linear
        regression).
    predictions_constant_data : str, a list of str
        Constants relevant to the model predictions (i.e. new x values in a linear
        regression).
    log_likelihood : dict of {str: str}, list of str or str, optional
        Pointwise log_likelihood for the data. log_likelihood is extracted from the
        posterior. It is recommended to use this argument as a dictionary whose keys
        are observed variable names and its values are the variables storing log
        likelihood arrays in the Stan code. In other cases, a dictionary with keys
        equal to its values is used. By default, if a variable ``log_lik`` is
        present in the Stan model, it will be retrieved as pointwise log
        likelihood values. Use ``False`` or set ``data.log_likelihood`` to
        false to avoid this behaviour.
    coords : dict[str, iterable]
        A dictionary containing the values that are used as index. The key
        is the name of the dimension, the values are the index values.
    dims : dict[str, list[str]]
        A mapping from variables to a list of coordinate names for the variable.
    posterior_model : stan.model.Model
        PyStan specific model object. Needed for automatic dtype parsing
        and for the extraction of observed data.
    prior_model : stan.model.Model
        PyStan specific model object. Needed for automatic dtype parsing.
    save_warmup : bool
        Save warmup iterations into InferenceData object. If not defined, use default
        defined by the rcParams.
    dtypes : dict
        A dictionary containing dtype information (int, float) for parameters.
        By default dtype information is extracted from the model code.
        Model code is extracted from model object.

    Returns
    -------
    DataTree
    """
    return PyStanConverter(
        posterior=posterior,
        posterior_model=posterior_model,
        posterior_predictive=posterior_predictive,
        predictions=predictions,
        prior=prior,
        prior_model=prior_model,
        prior_predictive=prior_predictive,
        observed_data=observed_data,
        constant_data=constant_data,
        predictions_constant_data=predictions_constant_data,
        log_likelihood=log_likelihood,
        coords=coords,
        dims=dims,
        save_warmup=save_warmup,
        dtypes=dtypes,
    ).to_datatree()



"""Helper functions to reorganize data."""

from numbers import Number

import numpy as np
import pandas as pd
import xarray as xr

from arviz_base.converters import convert_to_dataset
from arviz_base.labels import BaseLabeller
from arviz_base.rcparams import rcParams
from arviz_base.sel_utils import xarray_sel_iter
from arviz_base.utils import _var_names

__all__ = [
    "dataset_to_dataarray",
    "dataset_to_dataframe",
    "explode_dataset_dims",
    "extract",
    "references_to_dataset",
]


# TODO: remove this ignore about too many statements once the code uses validator functions
def extract(  # noqa: PLR0915
    data,
    group="posterior",
    sample_dims=None,
    *,
    combined=True,
    var_names=None,
    filter_vars=None,
    num_samples=None,
    weights=None,
    resampling_method=None,
    keep_dataset=False,
    random_seed=None,
):
    """Extract a group or group subset from a DataTree.

    Parameters
    ----------
    idata : DataTree-like
        DataTree from which to extract the data.
    group : str, optional
        Which group to extract data from.
    sample_dims : sequence of hashable, optional
        List of dimensions that should be considered sampling dimensions.
        Random subsets and potential stacking if ``combine=True`` happen
        over these dimensions only. Defaults to ``rcParams["data.sample_dims"]``.
    combined : bool, optional
        Combine `sample_dims` dimensions into ``sample``. Won't work if
        a dimension named ``sample`` already exists.
        It is irrelevant and ignored when `sample_dims` is a single dimension.
    var_names : str or list of str, optional
        Variables to be extracted. Prefix the variables by `~` when you want to exclude them.
    filter_vars : {None, "like", "regex"}, optional
        If `None` (default), interpret var_names as the real variables names. If "like",
        interpret var_names as substrings of the real variables names. If "regex",
        interpret var_names as regular expressions on the real variables names. A la
        `pandas.filter`.
        Like with plotting, sometimes it's easier to subset saying what to exclude
        instead of what to include
    num_samples : int, optional
        Extract only a subset of the samples. Only valid if ``combined=True`` or
        `sample_dims` represents a single dimension.
    weights : array-like, optional
        Extract a weighted subset of the samples. Only valid if `num_samples` is not ``None``.
    resampling_method : str, optional
        Method to use for resampling. Default is "multinomial". Options are "multinomial"
        and "stratified". For stratified resampling, weights must be provided.
        Default is "stratified" if weights are provided, "multinomial" otherwise.
    keep_dataset : bool, optional
        If true, always return a DataSet. If false (default) return a DataArray
        when there is a single variable.
    random_seed : int, numpy.Generator, optional
        Random number generator or seed. Only used if ``weights`` is not ``None``
        or if ``num_samples`` is not ``None``.

    Returns
    -------
    xarray.DataArray or xarray.Dataset

    Examples
    --------
    The default behaviour is to return the posterior group after stacking the chain and
    draw dimensions.

    .. jupyter-execute::

        import arviz_base as az
        idata = az.load_arviz_data("centered_eight")
        az.extract(idata)

    You can also indicate a subset to be returned, but in variables and in samples:

    .. jupyter-execute::

        az.extract(idata, var_names="theta", num_samples=100)

    To keep the chain and draw dimensions, use ``combined=False``.

    .. jupyter-execute::

        az.extract(idata, group="prior", combined=False)

    """
    # TODO: use validator function
    if sample_dims is None:
        sample_dims = rcParams["data.sample_dims"]
    if isinstance(sample_dims, str):
        sample_dims = [sample_dims]
    if len(sample_dims) == 1:
        combined = True
    if num_samples is not None and not combined:
        raise ValueError(
            "num_samples is only compatible with combined=True or length 1 sample_dims"
        )
    if weights is not None and num_samples is None:
        raise ValueError("weights are only compatible with num_samples")

    data = convert_to_dataset(data, group=group)
    var_names = _var_names(var_names, data, filter_vars)
    if var_names is not None:
        if len(var_names) == 1 and not keep_dataset:
            var_names = var_names[0]
        data = data[var_names]
    elif len(data.data_vars) == 1 and not keep_dataset:
        data = data[list(data.data_vars)[0]]

    if weights is not None:
        resampling_method = "stratified" if resampling_method is None else resampling_method
        weights = np.array(weights).ravel()
        if len(weights) != np.prod([data.sizes[dim] for dim in sample_dims]):
            raise ValueError("Weights must have the same size as `sample_dims`")
    else:
        resampling_method = "multinomial" if resampling_method is None else resampling_method

    if resampling_method not in ("multinomial", "stratified"):
        raise ValueError(f"Invalid resampling_method: {resampling_method}")

    if combined and len(sample_dims) != 1:
        data = data.stack(sample=sample_dims)
        combined_dim = "sample"
    elif len(sample_dims) == 1:
        combined_dim = sample_dims[0]

    if weights is not None or num_samples is not None:
        if random_seed is None:
            rng = np.random.default_rng()
        elif isinstance(random_seed, int | np.integer):
            rng = np.random.default_rng(random_seed)
        elif isinstance(random_seed, np.random.Generator):
            rng = random_seed
        else:
            raise ValueError(f"Invalid random_seed value: {random_seed}")

        replace = weights is not None

        if resampling_method == "multinomial":
            resample_indices = rng.choice(
                np.arange(data.sizes[combined_dim]),
                size=num_samples,
                p=weights,
                replace=replace,
            )
        elif resampling_method == "stratified":
            if weights is None:
                raise ValueError("Weights must be provided for stratified resampling")
            resample_indices = _stratified_resample(weights, rng)

        data = data.isel({combined_dim: resample_indices})

    return data


def _stratified_resample(weights, rng):
    """Stratified resampling."""
    N = len(weights)
    single_uniform = (rng.random(N) + np.arange(N)) / N
    indexes = np.zeros(N, dtype=int)
    cum_sum = np.cumsum(weights)

    i, j = 0, 0
    while i < N:
        if single_uniform[i] < cum_sum[j]:
            indexes[i] = j
            i += 1
        else:
            j += 1

    return indexes


def dataset_to_dataarray(
    ds, sample_dims=None, labeller=None, add_coords=True, new_dim="label", label_type="flat"
):
    """Convert a Dataset to a stacked DataArray, using a labeller to set coordinate values.

    Parameters
    ----------
    ds : Dataset
        Input data
    sample_dims : sequence of hashable, optional
        Dimensions that are present in all variables of `ds` and should be kept
        in the returned `DataArray`. All other variables will be stacked
        into `new_dim`.
    labeller : labeller, optional
        Labeller instance with a `make_label_flat` or `make_label_vert` method that
        will be use to generate the coordinate values along `new_dim`.
    add_coords : bool, default True
        Return multiple coordinate variables along `new_dim`. These will contain the newly
        generated labels, the stacked variable names, and stacked coordinate values.
    new_dim : hashable, default "label"
        Name of the new dimension that is created from stacking variables
        and dimensions not in `sample_dims`.
    label_type : {"flat", "vert"}, default "flat"
        if "flat", then `labeller.make_label_flat` method is used to generate the labels and if
        "vert", then `labeller.make_label_vert` method is used.

    Returns
    -------
    DataArray

    Examples
    --------
    Convert the posterior group into a stacked and labelled dataarray:

    .. jupyter-execute::

        import xarray as xr
        from arviz_base import load_arviz_data, dataset_to_dataarray
        xr.set_options(display_expand_data=False)

        idata = load_arviz_data("centered_eight")
        dataset_to_dataarray(idata.posterior.dataset)
    """
    if labeller is None:
        labeller = BaseLabeller()
    if sample_dims is None:
        sample_dims = rcParams["data.sample_dims"]

    if label_type not in ("flat", "vert"):
        raise ValueError(f"Invalid label_type: {label_type}")

    labeled_stack = ds.to_stacked_array(new_dim, sample_dims=sample_dims)
    labels = [
        (labeller.make_label_flat if label_type == "flat" else labeller.make_label_vert)(
            var_name, sel, isel
        )
        for var_name, sel, isel in xarray_sel_iter(ds, skip_dims=set(sample_dims))
    ]
    indexes = [
        idx_name
        for idx_name, idx in labeled_stack.xindexes.items()
        if (idx_name not in sample_dims) and (idx.dim not in sample_dims)
    ]
    labeled_stack = labeled_stack.drop_indexes(indexes).assign_coords({new_dim: labels})
    for idx_name in indexes:
        if idx_name == new_dim:
            continue
        if add_coords:
            labeled_stack = labeled_stack.set_xindex(idx_name)
        else:
            labeled_stack = labeled_stack.drop_vars(idx_name)
    return labeled_stack


def dataset_to_dataframe(ds, sample_dims=None, labeller=None, multiindex=False, new_dim="label"):
    """Convert a Dataset to a DataFrame via a stacked DataArray, using a labeller.

    Parameters
    ----------
    ds : Dataset
    sample_dims : sequence of hashable, optional
    labeller : labeller, optional
    multiindex : {"row", "column"} or bool, default False
    new_dim : hashable, default "label"

    Returns
    -------
    pandas.DataFrame

    Examples
    --------
    The output will have whatever that uses `sample_dims` as the columns of
    the DataFrame, so when these are much longer we might want to transpose the
    output:

    .. jupyter-execute::

        from arviz_base import load_arviz_data, dataset_to_dataframe
        idata = load_arviz_data("centered_eight")
        dataset_to_dataframe(idata.posterior.dataset)

    The default is to only return a single index, with the labels or tuples of coordinate
    values in the stacked dimensions. To keep all data from all coordinates as a multiindex
    use ``multiindex=True``

    .. jupyter-execute::

        dataset_to_dataframe(idata.posterior.dataset, multiindex=True)

    The only restriction on `sample_dims` is that it is present in all variables
    of the dataset. Consequently, we can compute statistical summaries,
    concatenate the results into a single dataset creating a new dimension.

    .. jupyter-execute::

        import xarray as xr

        dims = ["chain", "draw"]
        post = idata.posterior.dataset
        summaries = xr.concat(
            (
                post.mean(dims).expand_dims(summary=["mean"]),
                post.median(dims).expand_dims(summary=["median"]),
                post.quantile([.25, .75], dim=dims).rename(
                    quantile="summary"
                ).assign_coords(summary=["1st quartile", "3rd quartile"])
            ),
            dim="summary"
        )
        summaries

    Then convert the result into a DataFrame for ease of viewing.

    .. jupyter-execute::

        dataset_to_dataframe(summaries, sample_dims=["summary"]).T

    Note that if all summaries were scalar, it would not be necessary to use
    :meth:`~xarray.Dataset.expand_dims` or renaming dimensions, using
    :meth:`~xarray.Dataset.assign_coords` on the result to label the newly created
    dimension would be enough. But using this approach we already generate a dimension
    with coordinate values and can also combine non scalar summaries.
    """
    if sample_dims is None:
        sample_dims = rcParams["data.sample_dims"]
    da = dataset_to_dataarray(ds, sample_dims=sample_dims, labeller=labeller, new_dim=new_dim)
    sample_dim = sample_dims[0]
    if len(sample_dims) > 1:
        da = da.stack(sample=sample_dims)
        sample_dim = "sample"
    sample_idx = da[sample_dim]
    label_idx = da[new_dim]
    if multiindex is True or multiindex == "row":
        idx_dict = {
            idx_name: da[idx_name].to_numpy()
            for idx_name in da.xindexes
            if sample_dim in da[idx_name].dims
        }
        sample_idx = pd.MultiIndex.from_arrays(list(idx_dict.values()), names=list(idx_dict.keys()))
    if multiindex is True or multiindex == "column":
        idx_dict = {
            idx_name: da[idx_name].to_numpy()
            for idx_name in da.xindexes
            if new_dim in da[idx_name].dims
        }
        label_idx = pd.MultiIndex.from_arrays(list(idx_dict.values()), names=list(idx_dict.keys()))
    df = pd.DataFrame(
        da.transpose(sample_dim, new_dim).to_numpy(), columns=label_idx, index=sample_idx
    )
    return df


def explode_dataset_dims(ds, dim, labeller=None):
    """Explode dims of a dataset so each slice along them becomes its own variable.

    Parameters
    ----------
    ds : Dataset
    dim : hashable or sequence of hashable
        Dimension or dimensions along which slices to be stored as independent variables should
        be defined.
    labeller : labeller, optional
        Instance of a labeller class used to label the slices generated when exploding along `dim`.
        The method ``make_label_flat`` is used.

    Returns
    -------
    Dataset
        The dataset with all variables that have `dim` exploded into the respective slices
        as new variables.

    Examples
    --------
    In some cases, instead of ``theta`` as a ``(..., school)`` shape variable we'll want
    independent variables for each slice:

    .. jupyter-execute::

        from arviz_base import load_arviz_data, explode_dataset_dims
        import xarray as xr

        idata = load_arviz_data("centered_eight")
        explode_dataset_dims(idata.posterior.dataset, "school")
    """
    if isinstance(dim, str):
        dim = [dim]
    if labeller is None:
        labeller = BaseLabeller()
    return xr.Dataset(
        {
            labeller.make_label_flat(var_name, sel, isel): ds[var_name].sel(sel, drop=True)
            for var_name, sel, isel in xarray_sel_iter(
                ds, skip_dims={d for d in ds.dims if d not in dim}
            )
        }
    )


def references_to_dataset(references, ds, sample_dims=None, ref_dim=None):
    """Generate an :class:`~xarray.Dataset` compatible with `ds` from `references`.

    Cast common formats to provide references to a compatible Dataset.
    This function does not aim to be exhaustive, anything somewhat peculiar or complex
    will probably be better off building a Dataset manually instead.

    Parameters
    ----------
    references : scalar or 1D array-like or dict or DataArray or Dataset
        References to cast into a compatible dataset.

        * scalar inputs are interpreted as a reference line in each variable+coordinate not in
          `sample_dims` combination.
        * array-like inputs are interpreted as multiple reference lines in each variable+coordinate
          not in `sample_dims` combination. All subset having the same references
          and all references linked to every subset.
        * dict inputs are interpreted as array-like with each array matched to the variable
          corresponding to that dictionary key.
        * DataArray inputs are interpreted as an array-like if unnamed or as a single key
          dictionary if named.
        * Dataset inputs are returned as is but won't raise an error.

    ds : Dataset
        Dataset containing the data `references` should be compatible with.
    sample_dims : iterable of hashable, optional
        Sample dimensions in `ds`. The dimensions in the output will be the dimensions
        in `ds` minus `sample_dims` plus optionally a "ref_line_dim" for non-scalar references.
    ref_dim : str or list, optional
        Names for the new dimensions created during reference value broadcasting. Defaults to None.
        By default, "ref_dim" is added for 1D references and "ref_dim_x" for N-dimensional
        references when broadcasting over one or more variables.

    Returns
    -------
    Dataset
       A Dataset containing a subset of the variables, dimensions, and coordinate names from ds,
       with additional "ref_dim" dimensions added when multiple references are requested for one
       or more variables.

    See Also
    --------
    xarray.Dataset : Dataset constructor

    Examples
    --------
    Generate a reference dataset with 0 compatible with the centered eight example data:

    .. jupyter-execute::

        from arviz_base import load_arviz_data, references_to_dataset
        idata = load_arviz_data("centered_eight")
        references_to_dataset(0, idata.posterior.dataset)

    Generate a reference dataset with different references for each variable:

    .. jupyter-execute::

        references_to_dataset({"mu": -1, "tau": 1, "theta": 0}, idata.posterior.dataset)

    Or a similar case but with different number of references for each variable:

    .. jupyter-execute::

        ref_ds = references_to_dataset(
            {"mu": [-1, 0, 1], "tau": [1, 10], "theta": 0},
            idata.posterior.dataset
        )
        ref_ds

    Once we have a compatible dataset, we can for example compute the probability
    of the samples being above the reference value(s):

    .. jupyter-execute::

        (idata.posterior.dataset > ref_ds).mean()
    """
    # quick exit if dataset input
    if isinstance(references, xr.Dataset):
        return references
    # process argument defaults
    if sample_dims is None:
        sample_dims = rcParams["data.sample_dims"]
    if isinstance(sample_dims, str):
        sample_dims = [sample_dims]
    if isinstance(ref_dim, str):
        ref_dim = [ref_dim]

    # start covering cases, for dataarray, if its name is a variable convert to dataset
    # if it has no name treat is an array-like
    if isinstance(references, xr.DataArray):
        name = references.name
        if name is not None:
            if name not in ds.data_vars:
                raise ValueError(
                    "Input is a named DataArray whose name doesn't match any variable in `ds`. "
                    "Either use an unamed DataArray or ndarray or make sure the name matches."
                )
            return references.to_dataset()
        references = references.values
    # for scalars generate a dataset with requested shape full of reference value
    # check for numerical scalar following advise from
    # https://numpy.org/doc/2.2/reference/generated/numpy.isscalar.html
    if isinstance(references, Number):
        aux_ds = ds if sample_dims is None else ds.isel({dim: 0 for dim in sample_dims})
        return xr.full_like(aux_ds, references, dtype=np.array(references).dtype)
    # for array-like convert to dict so it is handled later on
    if isinstance(references, list | tuple | np.ndarray):
        references = {var_name: references for var_name in ds.data_vars}
    if isinstance(references, dict):
        ref_dict = {}
        for var_name, da in ds.items():
            if var_name not in references:
                continue
            ref_values = np.atleast_1d(references[var_name])
            new_dims = ref_values.shape
            if ref_dim is None:
                new_dim_names = (
                    ["ref_dim"]
                    if len(new_dims) == 1
                    else [f"ref_dim_{i}" for i in range(len(new_dims))]
                )
            else:
                if len(ref_dim) != len(new_dims):
                    raise ValueError(
                        f"ref_dim length ({len(ref_dim)}) does not match reference values "
                        f"length ({len(new_dims)}) for data variable {var_name}"
                    )
                new_dim_names = ref_dim[: len(new_dims)]
            sizes = {dim: length for dim, length in da.sizes.items() if dim not in sample_dims}
            full_shape = list(sizes.values()) + list(new_dims)
            data = np.broadcast_to(ref_values, full_shape)

            ref_dict[var_name] = xr.DataArray(
                data,
                dims=list(sizes) + new_dim_names,
                coords=dict(zip(new_dim_names, [np.arange(size) for size in new_dims]))
                | {
                    coord_name: coord_da
                    for coord_name, coord_da in da.coords.items()
                    if not set(coord_da.dims).intersection(sample_dims)
                },
            )

        return xr.Dataset(ref_dict)
    raise TypeError("Unrecognized input type for `references`")



# File generated with docstub

import locale
import logging
import os
import pprint
import re
import sys
import types
from collections.abc import Callable, Iterable, Iterator, MutableMapping
from pathlib import Path
from typing import Any, Literal, get_args

import numpy as np
from _typeshed import Incomplete

_log: Incomplete

ScaleKeyword: Incomplete

def _make_validate_choice(
    accepted_values: Iterable, allow_none: bool = ..., typeof: type = ...
) -> Callable | None: ...
def _make_validate_choice_regex(
    accepted_values: Iterable, accepted_values_regex: Iterable, allow_none: bool = ...
) -> Callable: ...
def _validate_positive_int(value) -> None: ...
def _validate_float(value) -> None: ...
def _validate_str(value) -> None: ...
def _validate_probability(value) -> float: ...
def _validate_rounding(value: int | str | None) -> str | int | None: ...
def _validate_boolean(value) -> None: ...
def _add_none_to_validator(base_validator) -> None: ...
def _validate_stats_module(value: str | types.ModuleType) -> str | types.ModuleType: ...
def _validate_bokeh_marker(value) -> None: ...
def _validate_backend(value: Incomplete) -> None: ...
def make_iterable_validator(
    scalar_validator, length=..., allow_none=..., allow_auto=...
) -> Callable: ...

_validate_float_or_none: Incomplete
_validate_positive_int_or_none: Incomplete
_validate_dims: Incomplete

defaultParams: Incomplete

class RcParams(MutableMapping):

    validate: Incomplete

    def __init__(self, *args, **kwargs) -> None: ...
    def __setitem__(self, key, val) -> None: ...
    def __getitem__(self, key) -> None: ...
    def __delitem__(self, key) -> None: ...
    def clear(self) -> None: ...
    def pop(self, key, default=...) -> None: ...
    def popitem(self) -> tuple[Any, Any]: ...
    def setdefault(self, key, default=...) -> None: ...
    def __repr__(self) -> str: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def find_all(self, pattern) -> None: ...
    def copy(self) -> None: ...

def get_arviz_rcfile() -> None: ...
def read_rcfile(fname) -> None: ...
def rc_params(ignore_files=...) -> None: ...

rcParams: Incomplete

class rc_context:
    def __init__(self, rc: dict | None = ..., fname: str | None = ...) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(self, exc_type, exc_value, exc_tb) -> None: ...



# File generated with docstub

import re
import warnings
from collections.abc import Hashable, Sequence
from typing import Literal

import numpy as np
from _typeshed import Incomplete
from numpy.typing import ArrayLike
from xarray import DataArray, Dataset

def _check_tilde_start(x: Incomplete) -> None: ...
def _var_names(
    var_names: str | list | None,
    data: Dataset | Sequence[Dataset],
    filter_vars: Literal[None, "like", "regex"] | None = ...,
    check_if_present: bool = ...,
) -> list | None: ...
def _subset_list(
    subset: str,
    whole_list: list,
    filter_items: Literal[None, "like", "regex"] | None = ...,
    warn=...,
    check_if_present=...,
) -> list | None: ...
def _get_coords(
    data: DataArray, coords: dict[Hashable, ArrayLike]
) -> Dataset | DataArray: ...
def expand_dims(x) -> None: ...



# File generated with docstub

import datetime
import importlib
import re
import types
import warnings
from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
from copy import deepcopy
from numbers import Number
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np
import xarray as xr
from _typeshed import Incomplete
from numpy.typing import ArrayLike, NDArray
from xarray import DataArray, Dataset

from arviz_base._version import __version__
from arviz_base.rcparams import rcParams

if TYPE_CHECKING:
    pass

RequiresArgTypeT: Incomplete
RequiresReturnTypeT: Incomplete

def generate_dims_coords(
    shape: Iterable[int],
    var_name: Hashable,
    dims: Iterable[Hashable] | None = ...,
    coords: dict[Any, ArrayLike] | None = ...,
    index_origin: int | None = ...,
    skip_event_dims: bool = ...,
    check_conventions: bool = ...,
) -> tuple[list[Hashable], dict[Any, NDArray]]: ...
def ndarray_to_dataarray(
    ary: Number | ArrayLike,
    var_name: Hashable,
    *,
    dims: Iterable[Hashable] | None = ...,
    sample_dims: Iterable[Hashable] | None = ...,
    coords: dict[Any, ArrayLike] | None = ...,
    index_origin: int | None = ...,
    skip_event_dims: bool = ...,
    check_conventions: bool = ...,
) -> DataArray: ...
def dict_to_dataset(
    data: Mapping[Any, ArrayLike],
    *,
    attrs: Mapping[Any, Any] | None = ...,
    inference_library: types.ModuleType | None = ...,
    coords: dict[Any, ArrayLike] | None = ...,
    dims: dict[Hashable, Sequence[Hashable]] | None = ...,
    sample_dims: Sequence[Hashable] | None = ...,
    index_origin: int | None = ...,
    skip_event_dims: bool = ...,
    check_conventions: bool = ...,
) -> Dataset: ...
def make_attrs(
    attrs: Mapping[Any, Any] | None = ...,
    inference_library: types.ModuleType | None = ...,
) -> dict: ...

class requires:
    def __init__(self, *props: str | list[str]) -> None: ...
    def __call__(
        self, func: Callable[[RequiresArgTypeT], RequiresReturnTypeT]
    ) -> Callable[[RequiresArgTypeT], RequiresReturnTypeT | None]: ...

def infer_stan_dtypes(stan_code) -> None: ...



# pylint: disable=wildcard-import,wrong-import-position
"""Base ArviZ features and converters."""

import logging

_log = logging.getLogger(__name__)

from arviz_base._version import __version__
from arviz_base.base import dict_to_dataset, generate_dims_coords, make_attrs, ndarray_to_dataarray
from arviz_base.citations import citations
from arviz_base.converters import convert_to_dataset, convert_to_datatree
from arviz_base.datasets import clear_data_home, get_data_home, list_datasets, load_arviz_data
from arviz_base.io_cmdstanpy import from_cmdstanpy
from arviz_base.io_dict import from_dict
from arviz_base.io_emcee import from_emcee
from arviz_base.io_numpyro import from_numpyro, from_numpyro_svi
from arviz_base.io_pystan import from_pystan
from arviz_base.rcparams import rc_context, rcParams
from arviz_base.reorg import (
    extract,
    dataset_to_dataarray,
    dataset_to_dataframe,
    explode_dataset_dims,
    references_to_dataset,
)
from arviz_base.sel_utils import xarray_sel_iter, xarray_var_iter, xarray_to_ndarray
from arviz_base.transform import get_unconstrained_samples
from arviz_base import testing, labels


__all__ = [
    "__version__",
    # base
    "citations",
    "dict_to_dataset",
    "generate_dims_coords",
    "make_attrs",
    "ndarray_to_dataarray",
    # converters
    "convert_to_dataset",
    "convert_to_datatree",
    # datasets
    "clear_data_home",
    "get_data_home",
    "list_datasets",
    "load_arviz_data",
    # io modules
    "from_cmdstanpy",
    "from_dict",
    "from_emcee",
    "from_numpyro",
    "from_numpyro_svi",
    # labels submodule
    "labels",
    # rcparams
    "rc_context",
    "rcParams",
    # reorg
    "extract",
    "dataset_to_dataarray",
    "dataset_to_dataframe",
    "explode_dataset_dims",
    "references_to_dataset",
    # sel_utils
    "xarray_sel_iter",
    "xarray_var_iter",
    "xarray_to_ndarray",
    # testing submodule
    "testing",
    # transform
    "get_unconstrained_samples",
]



"""ArviZ testing utilities."""

import numpy as np

from arviz_base import from_dict


def generate_base_data(seed=31):
    """Generate a base dataset for testing."""
    from scipy.stats import halfnorm, norm

    rng = np.random.default_rng(seed)
    mu = rng.normal(loc=1, size=(4, 100))
    tau = np.exp(rng.normal(size=(4, 100)))
    theta_orig = rng.uniform(size=7)
    theta = rng.normal(theta_orig[None, None, :], scale=1, size=(4, 100, 7))
    idxs = rng.choice(np.arange(7), size=29)
    x = np.linspace(0, 1, 29)
    obs = rng.normal(loc=x + theta_orig[idxs], scale=3)
    log_lik = norm(mu[:, :, None] * x[None, None, :] + theta[:, :, idxs], tau[:, :, None]).logpdf(
        obs[None, None, :]
    )
    log_lik = log_lik / log_lik.var()
    log_mu_prior = norm(0, 3).logpdf(mu)
    log_tau_prior = halfnorm(scale=5).logpdf(tau)
    log_theta_prior = norm(0, 1).logpdf(theta)
    prior_predictive = rng.normal(size=(1, 100, 29))
    posterior_predictive = rng.normal(size=(4, 100, 29))
    diverging = rng.choice([True, False], size=(4, 100), p=[0.1, 0.9])
    mu_prior = norm(0, 3).rvs(size=(1, 500), random_state=rng)
    tau_prior = halfnorm(0, 5).rvs(size=(1, 500), random_state=rng)
    theta_prior = norm(0, 1).rvs(size=(1, 500, 7), random_state=rng)
    energy = rng.normal(loc=50, scale=10, size=(4, 100))

    return {
        "posterior": {"mu": mu, "theta": theta, "tau": tau},
        "observed_data": {"y": obs},
        "log_likelihood": {"y": log_lik},
        "log_prior": {"mu": log_mu_prior, "theta": log_theta_prior, "tau": log_tau_prior},
        "prior_predictive": {"y": prior_predictive},
        "posterior_predictive": {"y": posterior_predictive},
        "sample_stats": {"diverging": diverging, "energy": energy},
        "prior": {"mu": mu_prior, "theta": theta_prior, "tau": tau_prior},
    }


def datatree(seed=31):
    """Generate a general DataTree."""
    base_data = generate_base_data(seed)

    dt = from_dict(
        base_data,
        dims={"theta": ["hierarchy"], "y": ["obs_dim"]},
    )
    dt["point_estimate"] = dt.posterior.mean(("chain", "draw"))
    dt["trunk"] = dt.azstats.eti(prob=0.5)
    dt["twig"] = dt.azstats.eti(prob=0.9)
    return dt


def datatree2(seed=17):
    """Generate a DataTree with a posterior and sample stats."""
    rng = np.random.default_rng(seed)
    mu = rng.normal(size=(4, 100))
    tau = rng.normal(size=(4, 100))
    theta = rng.normal(size=(4, 100, 7))
    theta_t = rng.normal(size=(4, 100, 7))
    diverging = rng.choice([True, False], size=(4, 100), p=[0.1, 0.9])

    return from_dict(
        {
            "posterior": {"mu": mu, "theta": theta, "tau": tau, "theta_t": theta_t},
            "sample_stats": {"diverging": diverging},
        },
        dims={"theta": ["hierarchy"], "theta_t": ["hierarchy"]},
    )


def datatree3(seed=17):
    """Generate a DataTree with discrete data."""
    rng = np.random.default_rng(seed)
    posterior_predictive = rng.poisson(4, size=(4, 100, 7))
    observed_data = rng.poisson(4, size=7)

    return from_dict(
        {
            "posterior_predictive": {"y": posterior_predictive},
            "observed_data": {"y": observed_data},
        },
        dims={"y": ["obs_dim"]},
    )


def datatree_binary(seed=17):
    """Generate a DataTree with binary data."""
    rng = np.random.default_rng(seed)
    posterior_predictive = rng.binomial(1, 0.5, size=(4, 100, 7))
    observed_data = rng.binomial(1, 0.5, size=7)
    log_likelihood = rng.normal(loc=0, scale=1, size=(4, 100, 7))

    return from_dict(
        {
            "posterior_predictive": {"y": posterior_predictive},
            "observed_data": {"y": observed_data},
            "log_likelihood": {"y": log_likelihood},
        },
        dims={"y": ["obs_dim"]},
    )


def datatree_regression(seed=17):
    """Generate a DataTree for regression data."""
    from scipy.stats import norm

    rng = np.random.default_rng(seed)
    n_obs = 100
    true_sigma = 0.9
    true_mu = 2 * np.linspace(-1, 1, n_obs)
    observed_data = true_mu + rng.normal(0, true_sigma, size=n_obs)

    posterior_sigma = rng.normal(true_sigma, 0.1, size=(4, 500))
    posterior_sigma = np.abs(posterior_sigma)

    posterior_mu = rng.normal(true_mu, true_sigma * 0.5, size=(4, 500, n_obs))
    posterior_predictive = rng.normal(posterior_mu, true_sigma, size=(4, 500, n_obs))
    log_likelihood = norm(posterior_mu, true_sigma).logpdf(observed_data)

    return from_dict(
        {
            "posterior": {"mu": posterior_mu, "sigma": posterior_sigma},
            "posterior_predictive": {"y": posterior_predictive},
            "observed_data": {"y": observed_data},
            "log_likelihood": {"y": log_likelihood},
        },
        dims={"y": ["obs_dim"]},
    )


def datatree_4d(seed=31):
    """Generate a DataTree with a 4D posterior."""
    rng = np.random.default_rng(seed)
    mu = rng.normal(size=(4, 100))
    theta = rng.normal(size=(4, 100, 5))
    eta = rng.normal(size=(4, 100, 5, 3))
    diverging = rng.choice([True, False], size=(4, 100), p=[0.1, 0.9])
    obs = rng.normal(size=(5, 3))
    prior_predictive = rng.normal(size=(1, 100, 5, 3))
    posterior_predictive = rng.normal(size=(4, 100, 5, 3))

    return from_dict(
        {
            "posterior": {"mu": mu, "theta": theta, "eta": eta},
            "observed_data": {"obs": obs},
            "prior_predictive": {"obs": prior_predictive},
            "posterior_predictive": {"obs": posterior_predictive},
            "sample_stats": {"diverging": diverging},
        },
        dims={"theta": ["hierarchy"], "eta": ["hierarchy", "group"]},
    )


def datatree_sample(seed=31):
    """Generate a DataTree with sample dimensions."""
    base_data = generate_base_data(seed)

    return from_dict(
        {
            group: {
                key: values[0] if group != "observed_data" else values
                for key, values in group_dict.items()
            }
            for group, group_dict in base_data.items()
        },
        dims={"theta": ["hierarchy"]},
        sample_dims=["sample"],
    )


def cmp():
    """Generate a comparison DataFrame."""
    import pandas as pd

    return pd.DataFrame(
        {
            "elpd": [-4.5, -14.3, -16.2],
            "p": [2.6, 2.3, 2.1],
            "elpd_diff": [0, 9.7, 11.3],
            "weight": [0.9, 0.1, 0],
            "se": [2.3, 2.7, 2.3],
            "dse": [0, 2.7, 2.3],
            "warning": [False, False, False],
        },
        index=["Model B", "Model A", "Model C"],
    )


def fake_dt():
    """Generate a fake prior/posterior DataTree."""
    rng = np.random.default_rng(42)

    return from_dict(
        {
            "posterior": {
                "a": rng.normal(size=(4, 100)),
                "b": rng.normal(size=(4, 100)),
            },
            "prior": {
                "a": rng.normal(size=(4, 100)),
                "b": rng.normal(size=(4, 100)),
            },
        }
    )


def datatree_censored(seed=42):
    """Create a sample DataTree with censored data."""
    rng = np.random.default_rng(seed)

    # Parameters
    n_obs = 100
    n_samples = 1000
    n_chains = 4
    observed_times = rng.exponential(scale=8, size=n_obs)
    status = rng.binomial(1, 0.7, size=n_obs)

    posterior_predictive_times = rng.exponential(scale=8, size=(n_chains, n_samples, n_obs))

    # Create DataTree
    return from_dict(
        {
            "observed_data": {
                "time": observed_times,
            },
            "constant_data": {
                "time": status,
            },
            "posterior_predictive": {
                "time": posterior_predictive_times,
            },
        },
        dims={"time": ["subject"]},
    )


def check_multiple_attrs(test_dict, parent):
    """Perform multiple hasattr checks on InferenceData objects.

    It is thought to first check if the parent object contains a given dataset,
    and then (if present) check the attributes of the dataset.

    Given the output of the function, all mismatches between expectation and reality can
    be retrieved: a single string indicates a group mismatch and a tuple of strings
    ``(group, var)`` indicates a mismatch in the variable ``var`` of ``group``.

    Parameters
    ----------
    test_dict : dict of {str : list of str}
        Its structure should be `{dataset1_name: [var1, var2], dataset2_name: [var]}`.
        A ``~`` at the beginning of a dataset or variable name indicates the name NOT
        being present must be asserted.
    parent : DataTree
        DataTree object on which to check the attributes.

    Returns
    -------
    list of (str or tuple of (str, str))
        List containing the failed checks. It will contain either the dataset_name or a
        tuple (dataset_name, var) for all non present attributes.

    Examples
    --------
    The output below indicates that ``posterior`` group was expected but not found, and
    variables ``a`` and ``b``:

        ["posterior", ("prior", "a"), ("prior", "b")]

    Another example could be the following:

        [("posterior", "a"), "~observed_data", ("sample_stats", "~log_likelihood")]

    In this case, the output indicates that variable ``a`` was not found in ``posterior``
    as it was expected, however, in the other two cases, the preceding ``~`` (kept from the
    input negation notation) indicates that ``observed_data`` group should not be present
    but was found in the InferenceData and that ``log_likelihood`` variable was found
    in ``sample_stats``, also against what was expected.

    """
    failed_attrs: list[str | tuple[str, str]] = []
    for dataset_name, attributes in test_dict.items():
        if dataset_name.startswith("~"):
            if hasattr(parent, dataset_name[1:]):
                failed_attrs.append(dataset_name)
        elif hasattr(parent, dataset_name):
            dataset = getattr(parent, dataset_name)
            for attribute in attributes:
                if attribute.startswith("~"):
                    if hasattr(dataset, attribute[1:]):
                        failed_attrs.append((dataset_name, attribute))
                elif not hasattr(dataset, attribute):
                    failed_attrs.append((dataset_name, attribute))
        else:
            failed_attrs.append(dataset_name)
    return failed_attrs



# File generated with docstub

from collections.abc import Hashable, Mapping, Sequence
from typing import TYPE_CHECKING, Any

from _typeshed import Incomplete

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray

CoordSpec: Incomplete
CoordOut: Incomplete
DimSpec: Incomplete
DimOut: Incomplete

DictData: Incomplete



"""How to cite ArviZ and its methods."""

import os
import re


def citations(methods=None, filepath=None, format_type="bibtex"):
    """
    List citations for ArviZ and the methods implemented in ArviZ.

    Parameters
    ----------
    methods : Callable or list of callable, optional
        Methods implemented in ArviZ from which to retrieve citations.
    filepath : str, optional
        Specifies the location to save the file with the citations.
        If ``None``, the result is returned as a string.
    format_type : str
       Specifies in which format the references will be displayed.
       Currently, only "bibtex" is supported.

    Examples
    --------
    >>> from arviz_base import citations
    >>> from arviz_stats import rhat
    >>> citations(methods=[rhat])  # Returns how to cite ArviZ and rhat
    >>> citations()  # Returns how to cite ArviZ
    """
    method_citations = [{"doi": "10.21105/joss.XXXXX"}]
    if methods is not None:
        if isinstance(methods, str):
            raise TypeError("you should pass an ArviZ function or list of functions.")
        if not isinstance(methods, list | tuple):
            methods = [methods]
        for method in methods:
            _extract_ids_per_entry(method_citations, method.__doc__)

    if format_type == "bibtex":
        header = _get_header(methods)
        citation_text = _find_bibtex_entries(header, method_citations)
        if filepath:
            with open(filepath, "w") as fw:
                fw.write(citation_text)
        else:
            return citation_text
    else:
        raise ValueError("Invalid value for format_type. Use 'bibtex'.")


def _extract_ids_per_entry(data, text):
    entries = re.split(r"\n\s*\.\. \[\d+\] ", text.strip())

    doi_pattern = re.compile(r"https?://doi\.org/(\S+)", re.IGNORECASE)
    url_pattern = re.compile(r"https?://(?!doi\.org)(\S+)", re.IGNORECASE)

    for entry in entries:
        doi_match = doi_pattern.search(entry)
        if doi_match:
            doi = doi_match.group(1).rstrip(".")
            data.append({"doi": doi})
        else:
            urls = [url.rstrip(".") for url in url_pattern.findall(entry)]
            if urls:
                data.append({"urls": urls})
    return data


def _find_bibtex_entries(header, data):
    ref_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "references.bib")
    with open(ref_path, encoding="utf-8") as fr:
        bibtex_text = fr.read()

    entries = re.split(r"\n(?=@)", bibtex_text)

    doi_field_pattern = re.compile(r'doi\s*=\s*[{"]([^}"]+)[}"]', re.IGNORECASE)
    url_field_pattern = re.compile(r'url\s*=\s*[{"]([^}"]+)[}"]', re.IGNORECASE)

    references = [header]
    for identifier in data:
        found_entry = ""
        for entry in entries:
            bib_dois = doi_field_pattern.findall(entry)
            bib_urls = url_field_pattern.findall(entry)

            if "doi" in identifier and any(identifier["doi"] in doi for doi in bib_dois):
                found_entry = entry.strip()
                break

            if "urls" in identifier and any(
                any(url in b_url or b_url in url for b_url in bib_urls)
                for url in identifier["urls"]
            ):
                found_entry = entry.strip()
                break
        if found_entry:
            if found_entry not in references:
                references.append(found_entry)

    return "\n\n".join(references)


def _get_header(methods=None):
    references = "Bibtex format citations for ArviZ paper\n"

    if methods is not None:
        methods_str = ", ".join([method.__name__ for method in methods])
        references = references.strip() + f", and\nfor the following methods: {methods_str}"

    return references



# File generated with docstub

from collections.abc import Hashable, Iterable, Sequence
from numbers import Number
from typing import Literal

import numpy as np
import pandas
import pandas as pd
import xarray as xr
from numpy.typing import ArrayLike
from xarray import DataArray, Dataset

from arviz_base.converters import convert_to_dataset
from arviz_base.labels import BaseLabeller
from arviz_base.rcparams import rcParams
from arviz_base.sel_utils import xarray_sel_iter
from arviz_base.utils import _var_names

from .labels import Labeller

__all__ = [
    "dataset_to_dataarray",
    "dataset_to_dataframe",
    "explode_dataset_dims",
    "extract",
    "references_to_dataset",
]

def extract(
    data,
    group: str = ...,
    sample_dims: Sequence[Hashable] | None = ...,
    *,
    combined: bool = ...,
    var_names: str | list[str] | None = ...,
    filter_vars: Literal[None, "like", "regex"] | None = ...,
    num_samples: int | None = ...,
    weights: ArrayLike | None = ...,
    resampling_method: str | None = ...,
    keep_dataset: bool = ...,
    random_seed: int | None = ...,
) -> DataArray | Dataset: ...
def _stratified_resample(weights, rng) -> None: ...
def dataset_to_dataarray(
    ds: Dataset,
    sample_dims: Sequence[Hashable] | None = ...,
    labeller: Labeller | None = ...,
    add_coords: bool = ...,
    new_dim: Hashable = ...,
    label_type: Literal["flat", "vert"] = ...,
) -> DataArray: ...
def dataset_to_dataframe(
    ds: Dataset,
    sample_dims: Sequence[Hashable] | None = ...,
    labeller: Labeller | None = ...,
    multiindex: Literal["row", "column"] | bool = ...,
    new_dim: Hashable = ...,
) -> pandas.DataFrame: ...
def explode_dataset_dims(
    ds: Dataset, dim: Hashable | Sequence[Hashable], labeller: Labeller | None = ...
) -> Dataset: ...
def references_to_dataset(
    references: Number | ArrayLike | dict | DataArray | Dataset,
    ds: Dataset,
    sample_dims: Iterable[Hashable] | None = ...,
    ref_dim: str | list | None = ...,
) -> Dataset: ...



# File generated with docstub

import logging
import re
from collections.abc import Iterable
from math import ceil

import numpy as np
import stan
from _typeshed import Incomplete
from xarray import DataArray, Dataset, DataTree

from arviz_base.base import (
    dict_to_dataset,
    generate_dims_coords,
    infer_stan_dtypes,
    make_attrs,
    requires,
)
from arviz_base.rcparams import rcParams

_log: Incomplete

class PyStanConverter:
    def __init__(
        self,
        *,
        posterior=...,
        posterior_model=...,
        posterior_predictive=...,
        predictions=...,
        prior=...,
        prior_model=...,
        prior_predictive=...,
        observed_data=...,
        constant_data=...,
        predictions_constant_data=...,
        log_likelihood=...,
        coords=...,
        dims=...,
        save_warmup=...,
        dtypes=...,
    ) -> None: ...
    def _warmup_return_to_dict(
        self,
        data: Incomplete,
        data_warmup: Incomplete,
        group: Incomplete,
        attrs: Incomplete = ...,
    ) -> None: ...
    def posterior_to_xarray(self) -> None: ...
    def sample_stats_to_xarray(self) -> None: ...
    def log_likelihood_to_xarray(self) -> None: ...
    def posterior_predictive_to_xarray(self) -> None: ...
    def predictions_to_xarray(self) -> None: ...
    def prior_to_xarray(self) -> None: ...
    def sample_stats_prior_to_xarray(self) -> None: ...
    def prior_predictive_to_xarray(self) -> None: ...
    def observed_data_to_xarray(self) -> None: ...
    def constant_data_to_xarray(self) -> None: ...
    def predictions_constant_data_to_xarray(self) -> None: ...
    def to_datatree(self) -> None: ...

def get_draws(
    fit, model=..., variables=..., ignore=..., warmup=..., dtypes=...
) -> None: ...
def get_sample_stats(
    fit, variables=..., ignore=..., warmup=..., dtypes=...
) -> None: ...
def get_attrs(fit, model=...) -> None: ...
def infer_dtypes(fit, model=...) -> None: ...
def from_pystan(
    posterior: stan.fit.Fit | None = ...,
    *,
    posterior_predictive: str | None = ...,
    predictions: str | None = ...,
    prior: stan.fit.Fit | None = ...,
    prior_predictive: str | None = ...,
    observed_data: str | None = ...,
    constant_data: str | None = ...,
    predictions_constant_data: str | None = ...,
    log_likelihood: dict[str, str] | None = ...,
    coords: dict[str, Iterable] | None = ...,
    dims: dict[str, list[str]] | None = ...,
    posterior_model: stan.model.Model | None = ...,
    prior_model: stan.model.Model | None = ...,
    save_warmup: bool | None = ...,
    dtypes: dict | None = ...,
) -> DataTree: ...



# File generated with docstub

from collections.abc import Callable, Generator, Hashable, Mapping, Sequence
from itertools import product, tee
from typing import Any

import numpy as np
import xarray as xr
from _typeshed import Incomplete
from numpy.typing import NDArray
from xarray import DataArray, Dataset

from arviz_base.labels import BaseLabeller
from arviz_base.rcparams import rcParams

__all__ = ["xarray_sel_iter", "xarray_var_iter", "xarray_to_ndarray"]

def _dims(data: Incomplete, var_name: Incomplete, skip_dims: Incomplete) -> None: ...
def _zip_dims(new_dims: Incomplete, vals: Incomplete) -> None: ...
def xarray_sel_iter(
    data: Dataset | DataArray,
    var_names: Sequence[Hashable] | None = ...,
    combined: bool | None = ...,
    skip_dims: set | None = ...,
    dim_to_idx: Mapping[Any, Hashable] | None = ...,
    reverse_selections: bool = ...,
) -> Generator[tuple[str, dict[Any, Any], dict[Any, Any]]]: ...
def xarray_var_iter(
    data: Dataset,
    var_names: Sequence[Hashable] | None = ...,
    combined: bool | None = ...,
    skip_dims: set | None = ...,
    dim_to_idx: Mapping[Any, Hashable] | None = ...,
    reverse_selections: bool = ...,
    dim_order: list | None = ...,
) -> Generator[tuple[str, dict[Any, Any], dict[Any, Any], DataArray]]: ...
def xarray_to_ndarray(
    data: Dataset,
    *,
    var_names: Sequence[Hashable] | None = ...,
    combined: bool = ...,
    label_fun: Callable | None = ...
) -> tuple[list, NDArray]: ...



# File generated with docstub

import hashlib
import itertools
import json
import os
import shutil
from collections import namedtuple
from urllib.request import urlretrieve

from _typeshed import Incomplete
from xarray import Dataset, open_datatree

from arviz_base.rcparams import rcParams

__all__ = ["get_data_home", "clear_data_home", "list_datasets", "load_arviz_data"]

LocalFileMetadata: Incomplete

RemoteFileMetadata: Incomplete
_EXAMPLE_DATA_DIR: Incomplete
_LOCAL_DATA_DIR: Incomplete

with open(os.path.join(_EXAMPLE_DATA_DIR, "data_local.json"), encoding="utf-8") as f:
    LOCAL_DATASETS: Incomplete

with open(os.path.join(_EXAMPLE_DATA_DIR, "data_remote.json"), encoding="utf-8") as f:
    REMOTE_DATASETS: Incomplete

def get_data_home(data_home: str | None = ...) -> None: ...
def clear_data_home(data_home: str | None = ...) -> None: ...
def _sha256(path) -> None: ...
def load_arviz_data(
    dataset: str | None = ..., data_home: str | None = ..., **kwargs: dict
) -> Dataset: ...
def list_datasets() -> list[str]: ...



# File generated with docstub

from collections.abc import Callable
from copy import deepcopy
from typing import Literal

import xarray as xr
from xarray import Dataset, DataTree

from arviz_base.utils import _var_names

def get_unconstrained_samples(
    idata: DataTree,
    var_names: str | list[str] | None = ...,
    filter_vars: Literal[None, "like", "regex"] | None = ...,
    transform_funcs: dict[str, Callable[[xr.DataArray], xr.DataArray]] | None = ...,
    group: str = ...,
    return_dataset: bool = ...,
) -> DataTree | Dataset: ...



"""ArviZ type definitions."""

from collections.abc import Hashable, Mapping, Sequence
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from numpy.typing import ArrayLike, NDArray

CoordSpec = Mapping[Any, "ArrayLike"]
CoordOut = dict[Any, "NDArray"]
DimSpec = Mapping[Any, Sequence[Hashable]]
DimOut = dict[Any, list[Hashable]]

DictData = Mapping[Any, "ArrayLike"]



# File generated with docstub

import warnings

import emcee
import numpy as np
from numpy.typing import ArrayLike
from xarray import DataTree

from arviz_base.base import dict_to_dataset
from arviz_base.rcparams import rc_context

def _verify_names(
    sampler: emcee.EnsembleSampler,
    var_names: list[str] | None,
    arg_names: list[str] | None,
    slices: list | None,
) -> list[str]: ...

class EmceeConverter:
    def __init__(
        self,
        sampler,
        var_names=...,
        slices=...,
        arg_names=...,
        arg_groups=...,
        blob_names=...,
        blob_groups=...,
        index_origin=...,
        coords=...,
        dims=...,
        check_conventions=...,
    ) -> None: ...
    def posterior_to_xarray(self) -> None: ...
    def args_to_xarray(self) -> None: ...
    def blobs_to_dict(self) -> None: ...
    def to_datatree(self) -> None: ...

def from_emcee(
    sampler: emcee.EnsembleSampler | None = ...,
    var_names: list[str] | None = ...,
    slices: list[ArrayLike | slice] | None = ...,
    arg_names: list[str] | None = ...,
    arg_groups: list[str] | None = ...,
    blob_names: list[str] | None = ...,
    blob_groups: list[str] | None = ...,
    index_origin=...,
    coords: dict[str, ArrayLike] | None = ...,
    dims: dict[str, list[str]] | None = ...,
    check_conventions: bool = ...,
) -> DataTree: ...



# File generated with docstub

from collections.abc import Hashable, Iterable, Mapping, Sequence
from typing import Any, Protocol

__all__ = [
    "mix_labellers",
    "Labeller",
    "BaseLabeller",
    "DimCoordLabeller",
    "IdxLabeller",
    "DimIdxLabeller",
    "MapLabeller",
    "NoVarLabeller",
]

class Labeller(Protocol):
    def make_label_vert(
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...
    def make_label_flat(
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...

def mix_labellers(labellers: Iterable[type], class_name: str = ...) -> type: ...

class BaseLabeller:
    def dim_coord_to_str(
        self, dim: Hashable, coord_val: Hashable, coord_idx: int | Sequence[int]
    ) -> str: ...
    def sel_to_str(
        self, sel: Mapping[Any, Hashable], isel: Mapping[Any, int | Sequence[int]]
    ) -> str: ...
    def var_name_to_str(self, var_name: str | None) -> str | None: ...
    def var_pp_to_str(
        self, var_name: str | None, pp_var_name: str | None
    ) -> str | None: ...
    def make_label_vert(
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...
    def make_label_flat(
        self,
        var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...
    def make_pp_label(
        self,
        var_name: str | None,
        pp_var_name: str | None,
        sel: Mapping[Any, Hashable],
        isel: Mapping[Any, int | Sequence[int]],
    ) -> str: ...

class DimCoordLabeller(BaseLabeller):
    def dim_coord_to_str(
        self, dim: Hashable, coord_val: Hashable, coord_idx: int | Sequence[int]
    ) -> str: ...

class IdxLabeller(BaseLabeller):
    def dim_coord_to_str(
        self, dim: Hashable, coord_val: Hashable, coord_idx: int | Sequence[int]
    ) -> str: ...

class DimIdxLabeller(BaseLabeller):
    def dim_coord_to_str(
        self, dim: Hashable, coord_val: Hashable, coord_idx: int | Sequence[int]
    ) -> str: ...

class MapLabeller(BaseLabeller):
    def __init__(
        self,
        var_name_map: Mapping[Hashable, Hashable] | None = ...,
        dim_map: Mapping[Hashable, Hashable] | None = ...,
        coord_map: Mapping[Hashable, Mapping[Hashable, Hashable]] | None = ...,
    ) -> None: ...
    def dim_coord_to_str(
        self, dim: Hashable, coord_val: Hashable, coord_idx: int | Sequence[int]
    ) -> str: ...
    def var_name_to_str(self, var_name: str | None) -> str | None: ...

class NoVarLabeller(BaseLabeller):
    def var_name_to_str(self, var_name: str | None) -> str | None: ...



# File generated with docstub

import warnings
from collections import defaultdict
from collections.abc import Callable
from typing import Any

import numpy as np
import numpyro
from _typeshed import Incomplete
from xarray import DataTree

from arviz_base.base import dict_to_dataset, requires
from arviz_base.rcparams import rc_context, rcParams
from arviz_base.utils import expand_dims

class SVIWrapper:
    def __init__(
        self,
        svi,
        *,
        svi_result,
        model_args=...,
        model_kwargs=...,
        num_samples: int = ...,
    ) -> None: ...
    def get_samples(self, seed=..., **kwargs) -> None: ...
    @property
    def sampler(self) -> None: ...
    def get_extra_fields(self, **kwargs) -> None: ...

def _add_dims(
    dims_a: dict[str, list[str]], dims_b: dict[str, list[str]]
) -> dict[str, list[str]]: ...
def infer_dims(
    model: Callable,
    model_args: tuple[Any, ...] | None = ...,
    model_kwargs: dict[str, Any] | None = ...,
) -> dict[str, list[str]]: ...

class NumPyroConverter:

    model: Incomplete
    nchains: Incomplete
    ndraws: Incomplete

    def __init__(
        self,
        *,
        posterior: numpyro.mcmc.MCMC | None = ...,
        prior: dict | None = ...,
        posterior_predictive: dict | None = ...,
        predictions: dict | None = ...,
        constant_data: dict | None = ...,
        predictions_constant_data: dict | None = ...,
        log_likelihood=...,
        index_origin: int | None = ...,
        coords: dict | None = ...,
        dims: dict[str, list[str]] | None = ...,
        pred_dims: dict | None = ...,
        extra_event_dims: dict | None = ...,
        num_chains: int = ...,
    ) -> None: ...
    def _get_model_trace(self, model, model_args, model_kwargs, key) -> None: ...
    def posterior_to_xarray(self) -> None: ...
    def sample_stats_to_xarray(self) -> None: ...
    def log_likelihood_to_xarray(self) -> None: ...
    def translate_posterior_predictive_dict_to_xarray(self, dct, dims) -> None: ...
    def posterior_predictive_to_xarray(self) -> None: ...
    def predictions_to_xarray(self) -> None: ...
    def priors_to_xarray(self) -> None: ...
    def observed_data_to_xarray(self) -> None: ...
    def constant_data_to_xarray(self) -> None: ...
    def predictions_constant_data_to_xarray(self) -> None: ...
    def to_datatree(self) -> None: ...
    def infer_dims(self) -> dict[str, list[str]]: ...
    def infer_pred_dims(self) -> dict[str, list[str]]: ...

def from_numpyro(
    posterior: numpyro.mcmc.MCMC | None = ...,
    *,
    prior: dict | None = ...,
    posterior_predictive: dict | None = ...,
    predictions: dict | None = ...,
    constant_data: dict | None = ...,
    predictions_constant_data: dict | None = ...,
    log_likelihood=...,
    index_origin: int | None = ...,
    coords: dict | None = ...,
    dims: dict[str, list[str]] | None = ...,
    pred_dims: dict | None = ...,
    extra_event_dims: dict | None = ...,
    num_chains: int = ...,
) -> DataTree: ...
def from_numpyro_svi(
    svi: numpyro.infer.svi.SVI,
    *,
    svi_result: numpyro.infer.svi.SVIRunResult,
    model_args: tuple | None = ...,
    model_kwargs: dict | None = ...,
    prior: dict | None = ...,
    posterior_predictive: dict | None = ...,
    predictions: dict | None = ...,
    constant_data: dict | None = ...,
    predictions_constant_data: dict | None = ...,
    log_likelihood=...,
    index_origin: int | None = ...,
    coords: dict | None = ...,
    dims: dict[str, list[str]] | None = ...,
    pred_dims: dict | None = ...,
    extra_event_dims: dict | None = ...,
    model=...,
    num_samples: int = ...,
) -> DataTree: ...



# File generated with docstub

import os
import re
from collections.abc import Callable

from _typeshed import Incomplete

def citations(
    methods: Callable | list[Callable] | None = ...,
    filepath: str | None = ...,
    format_type: str = ...,
) -> None: ...
def _extract_ids_per_entry(data: Incomplete, text: Incomplete) -> None: ...
def _find_bibtex_entries(header: Incomplete, data: Incomplete) -> None: ...
def _get_header(methods: Incomplete = ...) -> None: ...
